# Copyright (C) 2006, 2007, 2009, 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test the executable bit under various working tree formats."""

import os

from breezy import osutils
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestExecutable(TestCaseWithWorkingTree):

    def setUp(self):
        super().setUp()
        self.a_id = b"a-20051208024829-849e76f7968d7a86"
        self.b_id = b"b-20051208024829-849e76f7968d7a86"
        wt = self.make_branch_and_tree('b1')
        b = wt.branch
        tt = wt.transform()
        tt.new_file('a', tt.root, [b'a test\n'], self.a_id, True)
        tt.new_file('b', tt.root, [b'b test\n'], self.b_id, False)
        tt.apply()

        self.wt = wt

    def check_exist(self, tree):
        """Just check that both files have the right executable bits set"""
        tree.lock_read()
        self.assertTrue(tree.is_executable('a'), "'a' lost the execute bit")
        self.assertFalse(tree.is_executable('b'), "'b' gained an execute bit")
        tree.unlock()

    def check_empty(self, tree, ignore_inv=False):
        """Check that the files are truly missing
        :param ignore_inv: If you just delete files from a working tree
                the inventory still shows them, so don't assert that
                the inventory is empty, just that the tree doesn't have them
        """
        with tree.lock_read():
            if not ignore_inv and getattr(tree, 'root_inventory', None):
                self.assertEqual(
                    [('', tree.root_inventory.root)],
                    list(tree.root_inventory.iter_entries()))
            self.assertFalse(tree.has_filename('a'))
            self.assertFalse(tree.has_filename('b'))

    def commit_and_branch(self):
        """Commit the current tree, and create a second tree"""
        r1 = self.wt.commit('adding a,b')
        # Now make sure that 'bzr branch' also preserves the
        # executable bit
        dir2 = self.wt.branch.controldir.sprout('b2', revision_id=r1)
        wt2 = dir2.open_workingtree()
        self.assertEqual([r1], wt2.get_parent_ids())
        self.assertEqual(r1, wt2.branch.last_revision())
        return wt2, r1

    def test_01_is_executable(self):
        """Make sure that the tree was created and has the executable bit set"""
        self.check_exist(self.wt)

    def test_02_stays_executable(self):
        """reopen the tree and ensure it stuck."""
        self.wt = self.wt.controldir.open_workingtree()
        self.check_exist(self.wt)

    def test_03_after_commit(self):
        """Commit the change, and check the history"""
        r1 = self.wt.commit('adding a,b')

        rev_tree = self.wt.branch.repository.revision_tree(r1)
        self.check_exist(rev_tree)

    def test_04_after_removed(self):
        """Make sure reverting removed files brings them back correctly"""
        r1 = self.wt.commit('adding a,b')

        # Make sure the entries are gone
        os.remove('b1/a')
        os.remove('b1/b')
        self.check_empty(self.wt, ignore_inv=True)

        # Make sure that revert is able to bring them back,
        # and sets 'a' back to being executable

        rev_tree = self.wt.branch.repository.revision_tree(r1)

        self.wt.revert(['a', 'b'], rev_tree, backups=False)
        self.check_exist(self.wt)

    def test_05_removed_and_committed(self):
        """Check that reverting to an earlier commit restores them"""
        r1 = self.wt.commit('adding a,b')

        # Now remove them again, and make sure that after a
        # commit, they are still marked correctly
        os.remove('b1/a')
        os.remove('b1/b')
        r2 = self.wt.commit('removed')

        self.check_empty(self.wt)

        rev_tree = self.wt.branch.repository.revision_tree(r1)
        # Now revert back to the previous commit
        self.wt.revert(old_tree=rev_tree, backups=False)

        self.check_exist(self.wt)

    def test_06_branch(self):
        """branch b1=>b2 should preserve the executable bits"""
        # TODO: Maybe this should be a blackbox test
        wt2, r1 = self.commit_and_branch()

        self.check_exist(wt2)

    def test_07_pull(self):
        """Test that pull will handle bits correctly"""
        wt2, r1 = self.commit_and_branch()

        os.remove('b1/a')
        os.remove('b1/b')
        r2 = self.wt.commit('removed')

        # now wt2 can pull and the files should be removed

        # Make sure pull will delete the files
        wt2.pull(self.wt.branch)
        self.assertEqual([r2], wt2.get_parent_ids())
        self.assertEqual(r2, wt2.branch.last_revision())
        self.check_empty(wt2)

        # Now restore the files on the first branch and commit
        # so that the second branch can pull the changes
        # and make sure that the executable bit has been copied
        rev_tree = self.wt.branch.repository.revision_tree(r1)
        self.wt.revert(old_tree=rev_tree, backups=False)
        r3 = self.wt.commit('resurrected')

        self.check_exist(self.wt)

        wt2.pull(self.wt.branch)
        self.assertEqual([r3], wt2.get_parent_ids())
        self.assertEqual(r3, wt2.branch.last_revision())
        self.check_exist(wt2)

    def test_08_no_op_revert(self):
        """Just do a simple revert without anything changed

        The bits shouldn't swap.
        """
        r1 = self.wt.commit('adding a,b')
        rev_tree = self.wt.branch.repository.revision_tree(r1)
        self.wt.revert(old_tree=rev_tree, backups=False)
        self.check_exist(self.wt)

    def test_commit_with_exec_from_basis(self):
        self.wt._is_executable_from_path_and_stat = \
            self.wt._is_executable_from_path_and_stat_from_basis
        rev_id1 = self.wt.commit('one')
        rev_tree1 = self.wt.branch.repository.revision_tree(rev_id1)
        a_executable = rev_tree1.is_executable('a')
        b_executable = rev_tree1.is_executable('b')
        self.assertIsNot(None, a_executable)
        self.assertTrue(a_executable)
        self.assertIsNot(None, b_executable)
        self.assertFalse(b_executable)

    def test_use_exec_from_basis(self):
        self.wt._supports_executable = lambda: False
        self.addCleanup(self.wt.lock_read().unlock)
        self.assertTrue(self.wt.is_executable('a'))
        self.assertFalse(self.wt.is_executable('b'))



# Copyright (C) 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for InterBranch.fetch."""

from breezy.errors import FetchLimitUnsupported, NoRoundtrippingSupport
from breezy.revision import NULL_REVISION
from breezy.tests import TestNotApplicable
from breezy.tests.per_interbranch import TestCaseWithInterBranch


class TestInterBranchFetch(TestCaseWithInterBranch):

    def test_fetch_revisions(self):
        """Test fetch-revision operation."""
        wt = self.make_from_branch_and_tree('b1')
        b1 = wt.branch
        self.build_tree_contents([('b1/foo', b'hello')])
        wt.add(['foo'])
        rev1 = wt.commit('lala!', allow_pointless=False)

        b2 = self.make_to_branch('b2')
        try:
            b2.fetch(b1)
        except NoRoundtrippingSupport:
            raise TestNotApplicable(
                'lossless cross-vcs fetch %r to %r not supported' %
                (b1, b2))

        # fetch does not update the last revision
        self.assertEqual(NULL_REVISION, b2.last_revision())

        b2.repository.get_revision(rev1)
        tree = b2.repository.revision_tree(rev1)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(tree.get_file_text('foo'), b'hello')

    def test_fetch_revisions_limit(self):
        """Test fetch-revision operation."""
        builder = self.make_branch_builder('b1',
                                           format=self.branch_format_from._matchingcontroldir)
        builder.start_series()
        rev1 = builder.build_commit()
        rev2 = builder.build_commit()
        rev3 = builder.build_commit()
        builder.finish_series()
        b1 = builder.get_branch()
        b2 = self.make_to_branch('b2')
        try:
            b2.fetch(b1, limit=1)
        except FetchLimitUnsupported:
            raise TestNotApplicable(
                'interbranch does not support fetch limits')
        except NoRoundtrippingSupport:
            raise TestNotApplicable(
                'lossless cross-vcs fetch %r to %r not supported' %
                (b1, b2))

        # fetch does not update the last revision
        self.assertEqual(NULL_REVISION, b2.last_revision())

        self.assertEqual({rev1},
                         b2.repository.has_revisions([rev1, rev2, rev3]))

    def test_fetch_revisions_limit_incremental(self):
        """Test incremental fetch-revision operation with limit."""
        wt = self.make_from_branch_and_tree('b1')
        b1 = wt.branch
        self.build_tree_contents([('b1/foo', b'hello')])
        wt.add(['foo'])
        rev1 = wt.commit('lala!', allow_pointless=False)

        b2 = self.make_to_branch('b2')
        try:
            b2.fetch(b1, limit=1)
        except FetchLimitUnsupported:
            raise TestNotApplicable(
                'interbranch does not support fetch limits')
        except NoRoundtrippingSupport:
            raise TestNotApplicable(
                'lossless cross-vcs fetch %r to %r not supported' %
                (b1, b2))

        self.assertEqual(
            {rev1},
            b2.repository.has_revisions(
                [rev1, b'revision-2', b'revision-3']))

        rev2 = wt.commit('hmm')
        rev3 = wt.commit('hmmm')

        b2.fetch(b1, limit=1)

        # fetch does not update the last revision
        self.assertEqual(NULL_REVISION, b2.last_revision())

        self.assertEqual(
            {rev1, rev2},
            b2.repository.has_revisions([rev1, rev2, rev3]))



# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""InterTree implementation tests for bzr.

These test the conformance of all the InterTree variations to the expected API.
Specific tests for individual variations are in other places such as:
 - tests/test_workingtree.py
"""

import breezy
from breezy import revisiontree, tests
from breezy.bzr import inventorytree
from breezy.bzr.workingtree_3 import WorkingTreeFormat3
from breezy.bzr.workingtree_4 import WorkingTreeFormat4
from breezy.tests import default_transport, multiply_tests
from breezy.tests.per_tree import (TestCaseWithTree, return_parameter,
                                   revision_tree_from_workingtree)
from breezy.tree import InterTree


def return_provided_trees(test_case, source, target):
    """Return the source and target tree unaltered."""
    return source, target


class TestCaseWithTwoTrees(TestCaseWithTree):

    def not_applicable_if_cannot_represent_unversioned(self, tree):
        if isinstance(tree, revisiontree.RevisionTree):
            # The locked test trees conversion could not preserve the
            # unversioned file status. This is normal (e.g. InterDirstateTree
            # falls back to InterTree if the basis is not a
            # DirstateRevisionTree, and revision trees cannot have unversioned
            # files.
            raise tests.TestNotApplicable('cannot represent unversioned files')

    def not_applicable_if_missing_in(self, relpath, tree):
        if not tree.is_versioned(relpath):
            # The locked test trees conversion could not preserve the missing
            # file status. This is normal (e.g. InterDirstateTree falls back
            # to InterTree if the basis is not a DirstateRevisionTree, and
            # revision trees cannot have missing files.
            raise tests.TestNotApplicable('cannot represent missing files')

    def make_to_branch_and_tree(self, relpath):
        """Make a to_workingtree_format branch and tree."""
        made_control = self.make_controldir(relpath,
                                            format=self.workingtree_format_to._matchingcontroldir)
        made_control.create_repository()
        made_control.create_branch()
        return self.workingtree_format_to.initialize(made_control)


def make_scenarios(transport_server, transport_readonly_server, formats):
    """Transform the input formats to a list of scenarios.

    :param formats: A list of tuples:.
        (intertree_class,
         workingtree_format,
         workingtree_format_to,
         mutable_trees_to_test_trees)
    """
    result = []
    for (label, intertree_class,
         workingtree_format,
         workingtree_format_to,
         mutable_trees_to_test_trees) in formats:
        scenario = (label, {
            "transport_server": transport_server,
            "transport_readonly_server": transport_readonly_server,
            "bzrdir_format": workingtree_format._matchingcontroldir,
            "workingtree_format": workingtree_format,
            "intertree_class": intertree_class,
            "workingtree_format_to": workingtree_format_to,
            # mutable_trees_to_test_trees takes two trees and converts them to,
            # whatever relationship the optimiser under test requires.,
            "mutable_trees_to_test_trees": mutable_trees_to_test_trees,
            # workingtree_to_test_tree is set to disable changing individual,
            # trees: instead the mutable_trees_to_test_trees helper is used.,
            "_workingtree_to_test_tree": return_parameter,
            })
        result.append(scenario)
    return result


def mutable_trees_to_preview_trees(test_case, source, target):
    preview = target.preview_transform()
    test_case.addCleanup(preview.finalize)
    return source, preview.get_preview_tree()


def mutable_trees_to_revision_trees(test_case, source, target):
    """Convert both trees to repository based revision trees."""
    return (revision_tree_from_workingtree(test_case, source),
            revision_tree_from_workingtree(test_case, target))


def load_tests(loader, standard_tests, pattern):
    default_tree_format = WorkingTreeFormat3()
    submod_tests = loader.loadTestsFromModuleNames([
        'breezy.tests.per_intertree.test_compare',
        'breezy.tests.per_intertree.test_file_content_matches',
        'breezy.tests.per_intertree.test_find_path',
        ])
    test_intertree_permutations = [
        # test InterTree with two default-format working trees.
        (inventorytree.InterInventoryTree.__name__,
         inventorytree.InterInventoryTree,
         default_tree_format, default_tree_format,
         return_provided_trees)]
    for optimiser in InterTree.iter_optimisers():
        if optimiser is inventorytree.InterCHKRevisionTree:
            # XXX: we shouldn't use an Intertree object to detect inventories
            # -- vila 20090311
            chk_tree_format = WorkingTreeFormat4()
            chk_tree_format._get_matchingcontroldir = \
                lambda: breezy.controldir.format_registry.make_controldir('2a')
            test_intertree_permutations.append(
                (inventorytree.InterInventoryTree.__name__ + "(CHKInventory)",
                 inventorytree.InterInventoryTree,
                 chk_tree_format,
                 chk_tree_format,
                 mutable_trees_to_revision_trees))
        elif optimiser is breezy.bzr.workingtree_4.InterDirStateTree:
            # Its a little ugly to be conditional here, but less so than having
            # the optimiser listed twice.
            # Add once, compiled version
            test_intertree_permutations.append(
                (optimiser.__name__ + "(C)",
                 optimiser,
                 optimiser._matching_from_tree_format,
                 optimiser._matching_to_tree_format,
                 optimiser.make_source_parent_tree_compiled_dirstate))
            # python version
            test_intertree_permutations.append(
                (optimiser.__name__ + "(PY)",
                 optimiser,
                 optimiser._matching_from_tree_format,
                 optimiser._matching_to_tree_format,
                 optimiser.make_source_parent_tree_python_dirstate))
        elif (optimiser._matching_from_tree_format is not None and
              optimiser._matching_to_tree_format is not None):
            test_intertree_permutations.append(
                (optimiser.__name__,
                 optimiser,
                 optimiser._matching_from_tree_format,
                 optimiser._matching_to_tree_format,
                 optimiser._test_mutable_trees_to_test_trees))
    # PreviewTree does not have an InterTree optimiser class.
    test_intertree_permutations.append(
        (inventorytree.InterInventoryTree.__name__ + "(PreviewTree)",
         inventorytree.InterInventoryTree,
         default_tree_format,
         default_tree_format,
         mutable_trees_to_preview_trees))
    scenarios = make_scenarios(
        default_transport,
        # None here will cause a readonly decorator to be created
        # by the TestCaseWithTransport.get_readonly_transport method.
        None,
        test_intertree_permutations)
    # add the tests for the sub modules to the standard tests.
    return multiply_tests(submod_tests, scenarios, standard_tests)



# Copyright (C) 2009 by Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Revision pseudonyms."""

from collections import defaultdict

from breezy import errors, foreign, ui, urlutils


def parse_git_svn_id(text):
    """Parse a git svn id string.

    :param text: git svn id
    :return: URL, revision number, uuid
    """
    (head, uuid) = text.rsplit(" ", 1)
    (full_url, rev) = head.rsplit("@", 1)
    return (full_url.encode("utf-8"), int(rev), uuid.encode("utf-8"))


class SubversionBranchUrlFinder:

    def __init__(self):
        self._roots = defaultdict(set)

    def find_root(self, uuid, url):
        for root in self._roots[uuid]:
            if url.startswith(root):
                return root
        try:
            from subvertpy.ra import RemoteAccess
        except ModuleNotFoundError:
            return None
        c = RemoteAccess(url)
        root = c.get_repos_root()
        self._roots[uuid].add(root)
        return root

    def find_branch_path(self, uuid, url):
        root = self.find_root(uuid, url)
        if root is None:
            return None
        assert url.startswith(root)
        return url[len(root):].strip("/")


svn_branch_path_finder = SubversionBranchUrlFinder()


def _extract_converted_from_revid(rev):
    if "converted-from" not in rev.properties:
        return

    for line in rev.properties.get("converted-from", "").splitlines():
        (kind, serialized_foreign_revid) = line.split(" ", 1)
        yield (kind, serialized_foreign_revid)


def _extract_cscvs(rev):
    """Older-style launchpad-cscvs import."""
    if "cscvs-svn-branch-path" not in rev.properties:
        return
    yield (
        "svn", "{}:{}:{}".format(
            rev.properties["cscvs-svn-repository-uuid"],
            rev.properties["cscvs-svn-revision-number"],
            urlutils.quote(rev.properties["cscvs-svn-branch-path"].strip("/"))))


def _extract_git_svn_id(rev):
    if "git-svn-id" not in rev.properties:
        return
    (full_url, revnum, uuid) = parse_git_svn_id(rev.properties['git-svn-id'])
    branch_path = svn_branch_path_finder.find_branch_path(uuid, full_url)
    if branch_path is not None:
        yield ("svn", "%s:%d:%s" % (uuid, revnum, urlutils.quote(branch_path)))


def _extract_foreign_revision(rev):
    # Perhaps 'rev' is a foreign revision ?
    if getattr(rev, "foreign_revid", None) is not None:
        yield ("svn", rev.mapping.vcs.serialize_foreign_revid(rev.foreign_revid))


def _extract_foreign_revid(rev):
    # Try parsing the revision id
    try:
        foreign_revid, mapping = \
            foreign.foreign_vcs_registry.parse_revision_id(rev.revision_id)
    except errors.InvalidRevisionId:
        pass
    else:
        yield (
            mapping.vcs.abbreviation,
            mapping.vcs.serialize_foreign_revid(foreign_revid))


def _extract_debian_md5sum(rev):
    if 'deb-md5' in rev.properties:
        yield ("debian-md5sum", rev.properties["deb-md5"])


_foreign_revid_extractors = [
    _extract_converted_from_revid,
    _extract_cscvs,
    _extract_git_svn_id,
    _extract_foreign_revision,
    _extract_foreign_revid,
    _extract_debian_md5sum,
    ]


def extract_foreign_revids(rev):
    """Find ids of semi-equivalent revisions in foreign VCS'es.

    :param: Bazaar revision object
    :return: Set with semi-equivalent revisions.
    """
    ret = set()
    for extractor in _foreign_revid_extractors:
        ret.update(extractor(rev))
    return ret


def find_pseudonyms(repository, revids):
    """Find revisions that are pseudonyms of each other.

    :param repository: Repository object
    :param revids: Sequence of revision ids to check
    :return: Iterable over sets of pseudonyms
    """
    # Where have foreign revids ended up?
    conversions = defaultdict(set)
    # What are native revids conversions of?
    conversion_of = defaultdict(set)
    revs = repository.get_revisions(revids)
    pb = ui.ui_factory.nested_progress_bar()
    try:
        for i, rev in enumerate(revs):
            pb.update("finding pseudonyms", i, len(revs))
            for foreign_revid in extract_foreign_revids(rev):
                conversion_of[rev.revision_id].add(foreign_revid)
                conversions[foreign_revid].add(rev.revision_id)
    finally:
        pb.finished()
    done = set()
    for foreign_revid in conversions.keys():
        ret = set()
        check = set(conversions[foreign_revid])
        while check:
            x = check.pop()
            extra = set()
            for frevid in conversion_of[x]:
                extra.update(conversions[frevid])
                del conversions[frevid]
            del conversion_of[x]
            check.update(extra)
            ret.add(x)
        if len(ret) > 1:
            yield ret


def pseudonyms_as_dict(l):
    """Convert an iterable over pseudonyms to a dictionary.

    :param l: Iterable over sets of pseudonyms
    :return: Dictionary with pseudonyms for each revid.
    """
    ret = {}
    for pns in l:
        for pn in pns:
            ret[pn] = pns - {pn}
    return ret


def generate_rebase_map_from_pseudonyms(pseudonym_dict, existing, desired):
    """Create a rebase map from pseudonyms and existing/desired ancestry.

    :param pseudonym_dict: Dictionary with pseudonym as returned by
        pseudonyms_as_dict()
    :param existing: Existing ancestry, might need to be rebased
    :param desired: Desired ancestry
    :return: rebase map, as dictionary
    """
    rebase_map = {}
    for revid in existing:
        for pn in pseudonym_dict.get(revid, []):
            if pn in desired:
                rebase_map[revid] = pn
    return rebase_map



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black-box tests for brz handling non-ascii characters."""

import os
import sys

from breezy import osutils, tests, urlutils
from breezy.tests import EncodingAdapter
from breezy.tests.scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


class TestNonAscii(tests.TestCaseWithTransport):
    """Test that brz handles files/committers/etc which are non-ascii."""

    scenarios = EncodingAdapter.encoding_scenarios

    def setUp(self):
        super().setUp()
        self._check_can_encode_paths()

        self.overrideAttr(osutils, '_cached_user_encoding', self.encoding)
        email = self.info['committer'] + ' <joe@foo.com>'
        self.overrideEnv('BRZ_EMAIL', email)
        self.create_base()

    def run_bzr_decode(self, args, encoding=None, fail=False, retcode=None,
                       working_dir=None):
        """Run brz and decode the output into a particular encoding.

        Returns a string containing the stdout output from bzr.

        :param fail: If true, the operation is expected to fail with
            a UnicodeError.
        """
        if encoding is None:
            encoding = osutils.get_user_encoding()
        try:
            out = self.run_bzr_raw(args, encoding=encoding,
                                   retcode=retcode, working_dir=working_dir)[0]
            return out.decode(encoding)
        except UnicodeError as e:
            if not fail:
                raise
        else:
            # This command, run from the regular command line, will give a
            # traceback to the user.  That's not really good for a situation
            # that can be provoked just by the interaction of their input data
            # and locale, as some of these are.  What would be better?
            if fail:
                self.fail("Expected UnicodeError not raised")

    def _check_OSX_can_roundtrip(self, path, fs_enc=None):
        """Stop the test if it's about to fail or errors out.

        Until we get proper support on OSX for accented paths (in fact, any
        path whose NFD decomposition is different than the NFC one), this is
        the best way to keep test active (as opposed to disabling them
        completely). This is a stop gap. The tests should at least be rewritten
        so that the failing ones are clearly separated from the passing ones.
        """
        if fs_enc is None:
            fs_enc = sys.getfilesystemencoding()
        if sys.platform == 'darwin':
            encoded = path.encode(fs_enc)
            import unicodedata
            normal_thing = unicodedata.normalize('NFD', path)
            mac_encoded = normal_thing.encode(fs_enc)
            if mac_encoded != encoded:
                self.knownFailure(
                    'Unable to roundtrip path %r on OSX filesystem'
                    ' using encoding "%s"'
                    % (path, fs_enc))

    def _check_can_encode_paths(self):
        fs_enc = sys.getfilesystemencoding()
        terminal_enc = osutils.get_terminal_encoding()
        fname = self.info['filename']
        dir_name = self.info['directory']
        for thing in [fname, dir_name]:
            try:
                thing.encode(fs_enc)
            except UnicodeEncodeError:
                raise tests.TestSkipped(
                    'Unable to represent path %r in filesystem encoding "%s"'
                    % (thing, fs_enc))
            try:
                thing.encode(terminal_enc)
            except UnicodeEncodeError:
                raise tests.TestSkipped(
                    'Unable to represent path %r in terminal encoding "%s"'
                    ' (even though it is valid in filesystem encoding "%s")'
                    % (thing, terminal_enc, fs_enc))

    def create_base(self):
        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([('a', b'foo\n')])
        wt.add('a')
        wt.commit('adding a')

        self.build_tree_contents(
            [('b', b'non-ascii \xFF\xFF\xFC\xFB\x00 in b\n')])
        wt.add('b')
        wt.commit(self.info['message'])

        fname = self.info['filename']
        self.build_tree_contents([(fname, b'unicode filename\n')])
        wt.add(fname)
        wt.commit('And a unicode file\n')
        self.wt = wt
        # FIXME: We don't check that the add went well, in fact, it doesn't on
        # OSX (when LC_ALL is set correctly) because the added path doesn't
        # match the one used on OSX. But checking here will require more
        # invasive changes than adding the _check_OSX_can_roundtrip(), so I
        # punt for now -- vila 20090702

    def test_status(self):
        self.build_tree_contents(
            [(self.info['filename'], b'changed something\n')])
        txt = self.run_bzr_decode('status')
        self._check_OSX_can_roundtrip(self.info['filename'])
        self.assertEqual('modified:\n  {}\n'.format(self.info['filename']), txt)

        txt = self.run_bzr_decode('status', encoding='ascii')
        expected = 'modified:\n  {}\n'.format(
            self.info['filename'].encode('ascii', 'replace').decode('ascii'))
        self.assertEqual(expected, txt)

    def test_cat(self):
        # brz cat shouldn't change the contents
        # using run_brz since that doesn't decode
        txt = self.run_bzr_raw('cat b')[0]
        self.assertEqual(b'non-ascii \xFF\xFF\xFC\xFB\x00 in b\n', txt)

        self._check_OSX_can_roundtrip(self.info['filename'])
        txt = self.run_bzr_raw(['cat', self.info['filename']])[0]
        self.assertEqual(b'unicode filename\n', txt)

    def test_cat_revision(self):
        committer = self.info['committer']
        txt = self.run_bzr_decode('cat-revision -r 1')
        self.assertTrue(committer in txt,
                        'failed to find {!r} in {!r}'.format(committer, txt))

        msg = self.info['message']
        txt = self.run_bzr_decode('cat-revision -r 2')
        self.assertTrue(msg in txt, 'failed to find {!r} in {!r}'.format(msg, txt))

    def test_mkdir(self):
        txt = self.run_bzr_decode(['mkdir', self.info['directory']])
        self.assertEqual('added %s\n' % self.info['directory'], txt)

        # The text should be garbled, but the command should succeed
        txt = self.run_bzr_raw(['mkdir', self.info['directory'] + '2'],
                               encoding='ascii')[0]
        expected = 'added {}2\n'.format(self.info['directory'])
        expected = expected.encode('ascii', 'replace')
        self.assertEqual(expected, txt)

    def test_relpath(self):
        txt = self.run_bzr_decode(['relpath', self.info['filename']])
        self.assertEqual(self.info['filename'] + '\n', txt)

        self.run_bzr_decode(['relpath', self.info['filename']],
                            encoding='ascii', fail=True)

    def test_inventory(self):
        txt = self.run_bzr_decode('inventory')
        self._check_OSX_can_roundtrip(self.info['filename'])
        self.assertEqual(['a', 'b', self.info['filename']],
                         txt.splitlines())

        # inventory should fail if unable to encode
        self.run_bzr_decode('inventory', encoding='ascii', fail=True)

        # We don't really care about the ids themselves,
        # but the command shouldn't fail
        txt = self.run_bzr_decode('inventory --show-ids')

    def test_revno(self):
        # There isn't a lot to test here, since revno should always
        # be an integer
        self.assertEqual('3\n', self.run_bzr_decode('revno'))
        self.assertEqual('3\n', self.run_bzr_decode('revno', encoding='ascii'))

    def test_revision_info(self):
        self.run_bzr_decode('revision-info -r 1')

        # TODO: jam 20060105 If we support revisions with non-ascii characters,
        # this should be strict and fail.
        self.run_bzr_decode('revision-info -r 1', encoding='ascii')

    def test_mv(self):
        fname1 = self.info['filename']
        fname2 = self.info['filename'] + '2'
        dirname = self.info['directory']

        # fname1 already exists
        self.run_bzr_decode(['mv', 'a', fname1], fail=True)

        txt = self.run_bzr_decode(['mv', 'a', fname2])
        self.assertEqual('a => %s\n' % fname2, txt)
        self.assertPathDoesNotExist('a')
        self.assertPathExists(fname2)

        # After 'mv' we need to re-open the working tree
        self.wt = self.wt.controldir.open_workingtree()
        self.wt.commit('renamed to non-ascii')

        os.mkdir(dirname)
        self.wt.add(dirname)
        txt = self.run_bzr_decode(['mv', fname1, fname2, dirname])
        self._check_OSX_can_roundtrip(self.info['filename'])
        self.assertEqual(['{} => {}/{}'.format(fname1, dirname, fname1),
                          '{} => {}/{}'.format(fname2, dirname, fname2)], txt.splitlines())

        # The rename should still succeed
        newpath = '{}/{}'.format(dirname, fname2)
        txt = self.run_bzr_raw(['mv', newpath, 'a'], encoding='ascii')[0]
        self.assertPathExists('a')
        self.assertEqual(newpath.encode('ascii', 'replace') + b' => a\n', txt)

    def test_branch(self):
        # We should be able to branch into a directory that
        # has a unicode name, even if we can't display the name
        self.run_bzr_decode(['branch', '.', self.info['directory']])
        self.run_bzr_decode(['branch', '.', self.info['directory'] + '2'],
                            encoding='ascii')

    def test_pull(self):
        # Make sure we can pull from paths that can't be encoded
        dirname1 = self.info['directory']
        dirname2 = self.info['directory'] + '2'
        url1 = urlutils.local_path_to_url(dirname1)
        url2 = urlutils.local_path_to_url(dirname2)
        out_bzrdir = self.wt.controldir.sprout(url1)
        out_bzrdir.sprout(url2)

        self.build_tree_contents(
            [(osutils.pathjoin(dirname1, "a"), b'different text\n')])
        self.wt.commit('mod a')

        txt = self.run_bzr_decode('pull', working_dir=dirname2)

        expected = osutils.pathjoin(osutils.getcwd(), dirname1)
        self.assertEqual('Using saved parent location: %s/\n'
                         'No revisions or tags to pull.\n' % (expected,), txt)

        self.build_tree_contents(
            [(osutils.pathjoin(dirname1, 'a'), b'and yet more\n')])
        self.wt.commit('modifying a by ' + self.info['committer'])

        # We should be able to pull, even if our encoding is bad
        self.run_bzr_decode('pull --verbose', encoding='ascii',
                            working_dir=dirname2)

    def test_push(self):
        # TODO: Test push to an SFTP location
        # Make sure we can pull from paths that can't be encoded
        # TODO: jam 20060427 For drastically improving performance, we probably
        #       could create a local repository, so it wouldn't have to copy
        #       the files around as much.
        # Note that the tests don't actually fail, but if we don't set this
        # flag, we end up getting "Lock was not Unlocked" warnings

        dirname = self.info['directory']
        self.run_bzr_decode(['push', dirname])

        self.build_tree_contents([('a', b'adding more text\n')])
        self.wt.commit('added some stuff')

        # TODO: check the output text is properly encoded
        self.run_bzr_decode('push')

        self.build_tree_contents(
            [('a', b'and a bit more: \n%s\n' % (dirname.encode('utf-8'),))])

        self.wt.commit('Added some ' + dirname)
        self.run_bzr_decode('push --verbose', encoding='ascii')

        self.run_bzr_decode(['push', '--verbose', dirname + '2'])

        self.run_bzr_decode(['push', '--verbose', dirname + '3'],
                            encoding='ascii')

        self.run_bzr_decode(['push', '--verbose', '--create-prefix',
                             dirname + '4/' + dirname + '5'])
        self.run_bzr_decode(['push', '--verbose', '--create-prefix',
                             dirname + '6/' + dirname + '7'], encoding='ascii')

    def test_renames(self):
        fname = self.info['filename'] + '2'
        self.wt.rename_one('a', fname)
        txt = self.run_bzr_decode('renames')
        self.assertEqual('a => %s\n' % fname, txt)

        self.run_bzr_decode('renames', fail=True, encoding='ascii')

    def test_remove(self):
        fname = self.info['filename']
        txt = self.run_bzr_decode(['remove', fname], encoding='ascii')

    def test_remove_verbose(self):
        fname = self.info['filename']
        txt = self.run_bzr_decode(['remove', '--verbose', fname],
                                  encoding='ascii')

    def test_file_id(self):
        fname = self.info['filename']
        txt = self.run_bzr_decode(['file-id', fname])

        # TODO: jam 20060106 We don't support non-ascii file ids yet,
        #       so there is nothing which would fail in ascii encoding
        #       This *should* be retcode=3
        txt = self.run_bzr_decode(['file-id', fname], encoding='ascii')

    def test_file_path(self):
        # Create a directory structure
        fname = self.info['filename']
        dirname = self.info['directory']
        self.build_tree_contents([
            ('base/', ),
            (osutils.pathjoin('base', '{}/'.format(dirname)), )])
        self.wt.add('base')
        self.wt.add('base/' + dirname)
        path = osutils.pathjoin('base', dirname, fname)
        self._check_OSX_can_roundtrip(self.info['filename'])
        self.wt.rename_one(fname, path)
        self.wt.commit('moving things around')

        txt = self.run_bzr_decode(['file-path', path])

        # TODO: jam 20060106 We don't support non-ascii file ids yet,
        #       so there is nothing which would fail in ascii encoding
        #       This *should* be retcode=3
        txt = self.run_bzr_decode(['file-path', path], encoding='ascii')

    def test_revision_history(self):
        # TODO: jam 20060106 We don't support non-ascii revision ids yet,
        #       so there is nothing which would fail in ascii encoding
        txt = self.run_bzr_decode('revision-history')

    def test_ancestry(self):
        # TODO: jam 20060106 We don't support non-ascii revision ids yet,
        #       so there is nothing which would fail in ascii encoding
        txt = self.run_bzr_decode('ancestry')

    def test_diff(self):
        self._check_OSX_can_roundtrip(self.info['filename'])
        # TODO: jam 20060106 diff is a difficult one to test, because it
        #       shouldn't encode the file contents, but it needs some sort
        #       of encoding for the paths, etc which are displayed.
        self.build_tree_contents([(self.info['filename'], b'newline\n')])
        txt = self.run_bzr('diff', retcode=1)[0]

    def test_deleted(self):
        self._check_OSX_can_roundtrip(self.info['filename'])
        fname = self.info['filename']
        os.remove(fname)
        self.wt.remove(fname)

        txt = self.run_bzr_decode('deleted')
        self.assertEqual(fname + '\n', txt)

        txt = self.run_bzr_decode('deleted --show-ids')
        self.assertTrue(txt.startswith(fname))

        # Deleted should fail if cannot decode
        # Because it is giving the exact paths
        # which might be used by a front end
        self.run_bzr_decode('deleted', encoding='ascii', fail=True)

    def test_modified(self):
        fname = self.info['filename']
        self.build_tree_contents([(fname, b'modified\n')])

        txt = self.run_bzr_decode('modified')
        self._check_OSX_can_roundtrip(self.info['filename'])
        self.assertEqual('"' + fname + '"' + '\n', txt)

        self.run_bzr_decode('modified', encoding='ascii', fail=True)

    def test_added(self):
        fname = self.info['filename'] + '2'
        self.build_tree_contents([(fname, b'added\n')])
        self.wt.add(fname)

        txt = self.run_bzr_decode('added')
        self.assertEqual('"' + fname + '"' + '\n', txt)

        self.run_bzr_decode('added', encoding='ascii', fail=True)

    def test_root(self):
        dirname = self.info['directory']
        url = urlutils.local_path_to_url(dirname)
        self.run_bzr_decode('root')

        self.wt.controldir.sprout(url)

        txt = self.run_bzr_decode('root', working_dir=dirname)
        self.assertTrue(txt.endswith(dirname + '\n'))

        txt = self.run_bzr_decode('root', encoding='ascii', fail=True,
                                  working_dir=dirname)

    def test_log(self):
        fname = self.info['filename']

        txt = self.run_bzr_decode('log')
        self.assertNotEqual(-1, txt.find(self.info['committer']))
        self.assertNotEqual(-1, txt.find(self.info['message']))

        txt = self.run_bzr_decode('log --verbose')
        # FIXME: iso-8859-2 test shouldn't be skipped here --vila 20090702
        self._check_OSX_can_roundtrip(self.info['filename'])
        self.assertNotEqual(-1, txt.find(fname))

        # Make sure log doesn't fail even if we can't write out
        txt = self.run_bzr_raw('log --verbose', encoding='ascii')[0]
        self.assertNotEqual(-1, txt.find(fname.encode('ascii', 'replace')))

    def test_touching_revisions(self):
        fname = self.info['filename']
        txt = self.run_bzr_decode(['touching-revisions', fname])
        self._check_OSX_can_roundtrip(self.info['filename'])
        self.assertEqual('     3 added {}\n'.format(fname), txt)

        fname2 = self.info['filename'] + '2'
        self.wt.rename_one(fname, fname2)
        self.wt.commit('Renamed {} => {}'.format(fname, fname2))

        txt = self.run_bzr_decode(['touching-revisions', fname2])
        expected_txt = ('     3 added %s\n'
                        '     4 renamed %s => %s\n'
                        % (fname, fname, fname2))
        self.assertEqual(expected_txt, txt)

        self.run_bzr_decode(['touching-revisions', fname2], encoding='ascii',
                            fail=True)

    def test_ls(self):
        txt = self.run_bzr_decode('ls')
        self.assertEqual(sorted(['a', 'b', self.info['filename']]),
                         sorted(txt.splitlines()))
        txt = self.run_bzr_decode('ls --null')
        self.assertEqual(sorted(['', 'a', 'b', self.info['filename']]),
                         sorted(txt.split('\0')))

        txt = self.run_bzr_decode('ls', encoding='ascii', fail=True)
        txt = self.run_bzr_decode('ls --null', encoding='ascii', fail=True)

    def test_unknowns(self):
        fname = self.info['filename'] + '2'
        self.build_tree_contents([(fname, b'unknown\n')])

        # TODO: jam 20060112 brz unknowns is the only one which
        #       quotes paths do we really want it to?
        #       awilkins 20080521 added and modified do it now as well
        txt = self.run_bzr_decode('unknowns')
        self._check_OSX_can_roundtrip(self.info['filename'])
        self.assertEqual('"{}"\n'.format(fname), txt)

        self.run_bzr_decode('unknowns', encoding='ascii', fail=True)

    def test_ignore(self):
        fname2 = self.info['filename'] + '2.txt'
        self.build_tree_contents([(fname2, b'ignored\n')])

        def check_unknowns(expected):
            self.assertEqual(expected, list(self.wt.unknowns()))

        self._check_OSX_can_roundtrip(self.info['filename'])
        check_unknowns([fname2])

        self.run_bzr_decode(['ignore', './' + fname2])
        check_unknowns([])

        fname3 = self.info['filename'] + '3.txt'
        self.build_tree_contents([(fname3, b'unknown 3\n')])
        check_unknowns([fname3])

        # Ignore should not care what the encoding is
        # (right now it doesn't print anything)
        self.run_bzr_decode(['ignore', fname3], encoding='ascii')
        check_unknowns([])

        # Now try a wildcard match
        fname4 = self.info['filename'] + '4.txt'
        self.build_tree_contents([(fname4, b'unknown 4\n')])
        self.run_bzr_decode('ignore *.txt')
        check_unknowns([])

        # and a different wildcard that matches everything
        os.remove('.bzrignore')
        self.run_bzr_decode(['ignore', self.info['filename'] + '*'])
        check_unknowns([])

    def test_missing(self):
        # create empty tree as reference for missing
        self.make_branch_and_tree('empty-tree')

        msg = self.info['message']

        txt = self.run_bzr_decode('missing empty-tree')
        self.assertNotEqual(-1, txt.find(self.info['committer']))
        self.assertNotEqual(-1, txt.find(msg))

        # Make sure missing doesn't fail even if we can't write out
        txt = self.run_bzr_raw('missing empty-tree', encoding='ascii',
                               retcode=1)[0]
        self.assertNotEqual(-1, txt.find(msg.encode('ascii', 'replace')))

    def test_info(self):
        self.run_bzr_decode(['branch', '.', self.info['directory']])
        self.run_bzr_decode(['info', self.info['directory']])
        self.run_bzr_decode(['info', self.info['directory']],
                            encoding='ascii')

    def test_ignored(self):
        fname = self.info['filename'] + '1.txt'
        self.build_tree_contents([(fname, b'ignored\n')])
        self.run_bzr(['ignore', fname])
        txt = self.run_bzr_decode(['ignored'])
        self.assertEqual(txt, '%-50s %s\n' % (fname, fname))
        txt = self.run_bzr_decode(['ignored'], encoding='ascii')
        fname = fname.encode('ascii', 'replace').decode('ascii')
        self.assertEqual(txt, '%-50s %s\n' % (fname, fname))



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black-box tests for brz missing."""

from breezy import osutils, tests


class TestMissing(tests.TestCaseWithTransport):

    def assertMessages(self, out, must_have=(), must_not_have=()):
        """Check if commit messages are in or not in the output"""
        for m in must_have:
            self.assertContainsRe(out, r'\nmessage:\n  %s\n' % m)
        for m in must_not_have:
            self.assertNotContainsRe(out, r'\nmessage:\n  %s\n' % m)

    def test_missing_quiet(self):
        # <https://bugs.launchpad.net/bzr/+bug/284748>
        # create a source branch
        #
        # XXX: This still needs a test that missing is quiet when there are
        # missing revisions.
        a_tree = self.make_branch_and_tree('.')
        self.build_tree_contents([('a', b'initial\n')])
        a_tree.add('a')
        a_tree.commit(message='initial')

        out, err = self.run_bzr('missing -q .')
        self.assertEqual('', out)
        self.assertEqual('', err)

    def test_missing(self):
        missing_one = "You are missing 1 revision:"
        extra_one = "You have 1 extra revision:"

        # create a source branch
        a_tree = self.make_branch_and_tree('a')
        self.build_tree_contents([('a/a', b'initial\n')])
        a_tree.add('a')
        a_tree.commit(message='initial')

        # clone and add a differing revision
        b_tree = a_tree.controldir.sprout('b').open_workingtree()
        self.build_tree_contents([('b/a', b'initial\nmore\n')])
        b_tree.commit(message='more')

        def run_missing(args, retcode=1, working_dir=None):
            out, err = self.run_bzr(['missing'] + args,
                                    retcode=retcode, working_dir=working_dir)
            # we do not expect any error output.
            self.assertEqual('', err)
            return out.splitlines()

        def run_missing_a(args, retcode=1):
            return run_missing(['../a'] + args,
                               retcode=retcode, working_dir='b')

        def run_missing_b(args, retcode=1):
            return run_missing(['../b'] + args,
                               retcode=retcode, working_dir='a')

        # run missing in a against b
        # this should not require missing to take out a write lock on a
        # or b. So we take a write lock on both to test that at the same
        # time. This may let the test pass while the default branch is an
        # os-locking branch, but it will trigger failures with lockdir based
        # branches.
        a_branch = a_tree.branch
        a_branch.lock_write()
        b_branch = b_tree.branch
        b_branch.lock_write()

        lines = run_missing_b([])
        # we're missing the extra revision here
        self.assertEqual(missing_one, lines[0])
        # and we expect 8 lines of output which we trust at the moment to be
        # good.
        self.assertEqual(8, len(lines))
        # unlock the branches for the rest of the test
        a_branch.unlock()
        b_branch.unlock()

        # get extra revision from b
        a_tree.merge_from_branch(b_branch)
        a_tree.commit(message='merge')

        # compare again, but now we have the 'merge' commit extra
        lines = run_missing_b([])
        self.assertEqual(extra_one, lines[0])
        self.assertLength(8, lines)

        lines2 = run_missing_b(['--mine-only'])
        self.assertEqual(lines, lines2)

        lines3 = run_missing_b(['--theirs-only'], retcode=0)
        self.assertEqualDiff('Other branch has no new revisions.', lines3[0])

        # relative to a, missing the 'merge' commit
        lines = run_missing_a([])
        self.assertEqual(missing_one, lines[0])
        self.assertLength(8, lines)

        lines2 = run_missing_a(['--theirs-only'])
        self.assertEqual(lines, lines2)

        lines3 = run_missing_a(['--mine-only'], retcode=0)
        self.assertEqualDiff('This branch has no new revisions.', lines3[0])

        lines4 = run_missing_a(['--short'])
        self.assertLength(4, lines4)

        lines4a = run_missing_a(['-S'])
        self.assertEqual(lines4, lines4a)

        lines5 = run_missing_a(['--line'])
        self.assertLength(2, lines5)

        lines6 = run_missing_a(['--reverse'])
        self.assertEqual(lines6, lines)

        lines7 = run_missing_a(['--show-ids'])
        self.assertLength(11, lines7)

        lines8 = run_missing_a(['--verbose'])
        self.assertEqual("modified:", lines8[-2])
        self.assertEqual("  a", lines8[-1])

        self.assertEqualDiff('Other branch has no new revisions.',
                             run_missing_b(['--theirs-only'], retcode=0)[0])

        # after a pull we're back on track
        b_tree.pull(a_branch)
        self.assertEqualDiff("Branches are up to date.",
                             run_missing_b([], retcode=0)[0])
        self.assertEqualDiff('Branches are up to date.',
                             run_missing_a([], retcode=0)[0])
        # If you supply mine or theirs you only know one side is up to date
        self.assertEqualDiff('This branch has no new revisions.',
                             run_missing_a(['--mine-only'], retcode=0)[0])
        self.assertEqualDiff('Other branch has no new revisions.',
                             run_missing_a(['--theirs-only'], retcode=0)[0])

    def test_missing_filtered(self):
        # create a source branch
        a_tree = self.make_branch_and_tree('a')
        self.build_tree_contents([('a/a', b'initial\n')])
        a_tree.add('a')
        a_tree.commit(message='r1')
        # clone and add differing revisions
        b_tree = a_tree.controldir.sprout('b').open_workingtree()

        for i in range(2, 6):
            a_tree.commit(message='a%d' % i)
            b_tree.commit(message='b%d' % i)

        # local
        out, err = self.run_bzr('missing ../b --my-revision 3',
                                retcode=1, working_dir='a')
        self.assertMessages(out, ('a3', 'b2', 'b3', 'b4', 'b5'), ('a2', 'a4'))

        out, err = self.run_bzr('missing ../b --my-revision 3..4',
                                retcode=1, working_dir='a')
        self.assertMessages(out, ('a3', 'a4'), ('a2', 'a5'))

        # remote
        out, err = self.run_bzr('missing ../b -r 3',
                                retcode=1, working_dir='a')
        self.assertMessages(out, ('a2', 'a3', 'a4', 'a5', 'b3'), ('b2', 'b4'))

        out, err = self.run_bzr('missing ../b -r 3..4',
                                retcode=1, working_dir='a')
        self.assertMessages(out, ('b3', 'b4'), ('b2', 'b5'))

        # both
        out, err = self.run_bzr('missing ../b --my-revision 3..4 -r 3..4',
                                retcode=1, working_dir='a')
        self.assertMessages(out, ('a3', 'a4', 'b3', 'b4'),
                            ('a2', 'a5', 'b2', 'b5'))

    def test_missing_check_last_location(self):
        # check that last location shown as filepath not file URL

        # create a source branch
        wt = self.make_branch_and_tree('a')
        b = wt.branch
        self.build_tree(['a/foo'])
        wt.add('foo')
        wt.commit('initial')

        location = osutils.getcwd() + '/a/'

        # clone
        b.controldir.sprout('b')

        # check last location
        lines, err = self.run_bzr('missing', working_dir='b')
        self.assertEqual('Using saved parent location: %s\n'
                         'Branches are up to date.\n' % location,
                         lines)
        self.assertEqual('', err)

    def test_missing_directory(self):
        """Test --directory option"""

        # create a source branch
        a_tree = self.make_branch_and_tree('a')
        self.build_tree_contents([('a/a', b'initial\n')])
        a_tree.add('a')
        a_tree.commit(message='initial')

        # clone and add a differing revision
        b_tree = a_tree.controldir.sprout('b').open_workingtree()
        self.build_tree_contents([('b/a', b'initial\nmore\n')])
        b_tree.commit(message='more')

        out2, err2 = self.run_bzr('missing --directory a b', retcode=1)
        out1, err1 = self.run_bzr('missing ../b', retcode=1, working_dir='a')
        self.assertEqualDiff(out1, out2)
        self.assertEqualDiff(err1, err2)

    def test_missing_tags(self):
        """Test showing tags"""

        # create a source branch
        a_tree = self.make_branch_and_tree('a')
        self.build_tree_contents([('a/a', b'initial\n')])
        a_tree.add('a')
        a_tree.commit(message='initial')

        # clone and add a differing revision
        b_tree = a_tree.controldir.sprout('b').open_workingtree()
        self.build_tree_contents([('b/a', b'initial\nmore\n')])
        b_tree.commit(message='more')
        b_tree.branch.tags.set_tag('a-tag', b_tree.last_revision())

        for log_format in ['long', 'short', 'line']:
            out, err = self.run_bzr(
                f'missing --log-format={log_format} ../a',
                working_dir='b', retcode=1)
            self.assertContainsString(out, 'a-tag')

            out, err = self.run_bzr(
                f'missing --log-format={log_format} ../b',
                working_dir='a', retcode=1)
            self.assertContainsString(out, 'a-tag')



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests specific to Repository implementations that use foreign VCS'es."""


from breezy.tests import TestCase, TestCaseWithTransport


class TestRepositoryFormat(TestCase):

    def test_format_string(self):
        self.assertRaises(NotImplementedError,
                          self.repository_format.get_format_string)

    def test_network_name(self):
        self.assertIsInstance(self.repository_format.network_name(),
                              bytes)

    def test_format_description(self):
        self.assertIsInstance(self.repository_format.get_format_description(),
                              str)


class ForeignRepositoryFactory:
    """Factory of repository for ForeignRepositoryTests."""

    def make_repository(self, transport):
        """Create a new, valid, repository. May or may not contain
        data."""
        raise NotImplementedError(self.make_repository)


class ForeignRepositoryTests(TestCaseWithTransport):
    """Basic tests for foreign repository implementations.

    These tests mainly make sure that the implementation covers the required
    bits of the API and returns semi-reasonable values, that are
    at least of the expected types and in the expected ranges.
    """

    # XXX: Some of these tests could be moved into a common testcase for
    # both native and foreign repositories.

    # Set to an instance of ForeignRepositoryFactory by the scenario
    repository_factory = None

    def make_repository(self):
        return self.repository_factory.make_repository(self.get_transport())

    def test_make_working_trees(self):
        """Test that Repository.make_working_trees() returns a boolean."""
        repo = self.make_repository()
        self.assertIsInstance(repo.make_working_trees(), bool)

    def test_get_physical_lock_status(self):
        """Test that a new repository is not locked by default."""
        repo = self.make_repository()
        self.assertFalse(repo.get_physical_lock_status())

    def test_is_shared(self):
        """Test that is_shared() returns a bool."""
        repo = self.make_repository()
        self.assertIsInstance(repo.is_shared(), bool)

    def test_gather_stats(self):
        """Test that gather_stats() will at least return a dictionary
        with the required keys."""
        repo = self.make_repository()
        stats = repo.gather_stats()
        self.assertIsInstance(stats, dict)



# Copyright (C) 2006-2009, 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import conflicts, errors, osutils, revisiontree, tests
from breezy import transport as _mod_transport
from breezy.bzr import workingtree_4
from breezy.tests import TestSkipped
from breezy.tests.per_tree import TestCaseWithTree
from breezy.tree import MissingNestedTree


class TestAnnotate(TestCaseWithTree):

    def test_annotate(self):
        work_tree = self.make_branch_and_tree('wt')
        tree = self.get_tree_no_parents_abc_content(work_tree)
        tree_revision = getattr(tree, 'get_revision_id', lambda: b'current:')()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        for revision, line in tree.annotate_iter('a'):
            self.assertEqual(b'contents of a\n', line)
            self.assertEqual(tree_revision, revision)
        tree_revision = getattr(tree, 'get_revision_id', lambda: b'random:')()
        for revision, line in tree.annotate_iter('a', default_revision=b'random:'):
            self.assertEqual(b'contents of a\n', line)
            self.assertEqual(tree_revision, revision)


class TestPlanFileMerge(TestCaseWithTree):

    def test_plan_file_merge(self):
        work_a = self.make_branch_and_tree('wta')
        self.build_tree_contents([('wta/file', b'a\nb\nc\nd\n')])
        work_a.add('file')
        work_a.commit('base version')
        work_b = work_a.controldir.sprout('wtb').open_workingtree()
        self.build_tree_contents([('wta/file', b'b\nc\nd\ne\n')])
        tree_a = self.workingtree_to_test_tree(work_a)
        if getattr(tree_a, 'plan_file_merge', None) is None:
            raise tests.TestNotApplicable(
                'Tree does not support plan_file_merge')
        tree_a.lock_read()
        self.addCleanup(tree_a.unlock)
        self.build_tree_contents([('wtb/file', b'a\nc\nd\nf\n')])
        tree_b = self.workingtree_to_test_tree(work_b)
        tree_b.lock_read()
        self.addCleanup(tree_b.unlock)
        self.assertEqual([
            ('killed-a', b'a\n'),
            ('killed-b', b'b\n'),
            ('unchanged', b'c\n'),
            ('unchanged', b'd\n'),
            ('new-a', b'e\n'),
            ('new-b', b'f\n'),
        ], list(tree_a.plan_file_merge('file', tree_b)))


class TestReference(TestCaseWithTree):

    def skip_if_no_reference(self, tree):
        if not tree.supports_tree_reference():
            raise tests.TestNotApplicable('Tree references not supported')

    def create_nested(self):
        work_tree = self.make_branch_and_tree('wt')
        with work_tree.lock_write():
            self.skip_if_no_reference(work_tree)
            subtree = self.make_branch_and_tree('wt/subtree')
            subtree.commit('foo')
            work_tree.add_reference(subtree)
        tree = self._convert_tree(work_tree)
        self.skip_if_no_reference(tree)
        return tree, subtree

    def test_get_reference_revision(self):
        tree, subtree = self.create_nested()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(
            subtree.last_revision(),
            tree.get_reference_revision('subtree'))

    def test_iter_references(self):
        tree, subtree = self.create_nested()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(
            ['subtree'],
            list(tree.iter_references()))

    def test_get_nested_tree(self):
        tree, subtree = self.create_nested()
        try:
            changes = subtree.changes_from(tree.get_nested_tree('subtree'))
            self.assertFalse(changes.has_changed())
        except MissingNestedTree:
            # Also okay.
            pass

    def test_get_root_id(self):
        # trees should return some kind of root id; it can be none
        tree = self.make_branch_and_tree('tree')
        if not tree.supports_file_ids:
            raise tests.TestNotApplicable('file ids not supported')
        root_id = tree.path2id('')
        if root_id is not None:
            self.assertIsInstance(root_id, bytes)

    def test_is_versioned(self):
        tree = self.make_branch_and_tree('tree')
        self.assertTrue(tree.is_versioned(''))
        self.assertFalse(tree.is_versioned('blah'))
        self.build_tree(['tree/dir/', 'tree/dir/file'])
        self.assertFalse(tree.is_versioned('dir'))
        self.assertFalse(tree.is_versioned('dir/'))
        tree.add(['dir', 'dir/file'])
        self.assertTrue(tree.is_versioned('dir'))
        self.assertTrue(tree.is_versioned('dir/'))


class TestFileIds(TestCaseWithTree):

    def setUp(self):
        super().setUp()
        if not self.workingtree_format.supports_setting_file_ids:
            raise tests.TestNotApplicable("working tree does not support setting file ids")

    def test_id2path(self):
        # translate from file-id back to path
        work_tree = self.make_branch_and_tree('wt')
        tree = self.get_tree_no_parents_abc_content(work_tree)
        a_id = tree.path2id('a')
        with tree.lock_read():
            self.assertEqual('a', tree.id2path(a_id))
            # other ids give an error- don't return None for this case
            self.assertRaises(errors.NoSuchId, tree.id2path, b'a')

    def test_all_file_ids(self):
        work_tree = self.make_branch_and_tree('wt')
        tree = self.get_tree_no_parents_abc_content(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(tree.all_file_ids(),
                         {tree.path2id('a'), tree.path2id(''),
                          tree.path2id('b'), tree.path2id('b/c')})


class TestStoredKind(TestCaseWithTree):

    def test_stored_kind(self):
        tree = self.make_branch_and_tree('tree')
        work_tree = self.make_branch_and_tree('wt')
        tree = self.get_tree_no_parents_abc_content(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual('file', tree.stored_kind('a'))
        self.assertEqual('directory', tree.stored_kind('b'))


class TestFileContent(TestCaseWithTree):

    def test_get_file(self):
        work_tree = self.make_branch_and_tree('wt')
        tree = self.get_tree_no_parents_abc_content_2(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        # Test lookup without path works
        file_without_path = tree.get_file('a')
        try:
            lines = file_without_path.readlines()
            self.assertEqual([b'foobar\n'], lines)
        finally:
            file_without_path.close()

    def test_get_file_context_manager(self):
        work_tree = self.make_branch_and_tree('wt')
        tree = self.get_tree_no_parents_abc_content_2(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        with tree.get_file('a') as f:
            self.assertEqual(b'foobar\n', f.read())

    def test_get_file_text(self):
        work_tree = self.make_branch_and_tree('wt')
        tree = self.get_tree_no_parents_abc_content_2(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        # test read by path
        self.assertEqual(b'foobar\n', tree.get_file_text('a'))

    def test_get_file_lines(self):
        work_tree = self.make_branch_and_tree('wt')
        tree = self.get_tree_no_parents_abc_content_2(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        # test read by path
        self.assertEqual([b'foobar\n'], tree.get_file_lines('a'))

    def test_get_file_lines_multi_line_breaks(self):
        work_tree = self.make_branch_and_tree('wt')
        self.build_tree_contents([('wt/foobar', b'a\rb\nc\r\nd')])
        work_tree.add('foobar')
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([b'a\rb\n', b'c\r\n', b'd'],
                         tree.get_file_lines('foobar'))


class TestExtractFilesBytes(TestCaseWithTree):

    def test_iter_files_bytes(self):
        work_tree = self.make_branch_and_tree('wt')
        self.build_tree_contents([('wt/foo', b'foo'),
                                  ('wt/bar', b'bar'),
                                  ('wt/baz', b'baz')])
        work_tree.add(['foo', 'bar', 'baz'])
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        extracted = {i: b''.join(b) for i, b in
                         tree.iter_files_bytes([('foo', 'id1'),
                                                ('bar', 'id2'),
                                                ('baz', 'id3')])}
        self.assertEqual(b'foo', extracted['id1'])
        self.assertEqual(b'bar', extracted['id2'])
        self.assertEqual(b'baz', extracted['id3'])
        self.assertRaises(_mod_transport.NoSuchFile, lambda: list(
                          tree.iter_files_bytes(
                              [('qux', 'file1-notpresent')])))


class TestConflicts(TestCaseWithTree):

    def test_conflicts(self):
        """Tree.conflicts() should return a ConflictList instance."""
        work_tree = self.make_branch_and_tree('wt')
        tree = self._convert_tree(work_tree)
        self.assertIsInstance(tree.conflicts(), conflicts.ConflictList)


class TestIterEntriesByDir(TestCaseWithTree):

    def test_iteration_order(self):
        work_tree = self.make_branch_and_tree('.')
        self.build_tree(['a/', 'a/b/', 'a/b/c', 'a/d/', 'a/d/e', 'f/', 'f/g'])
        work_tree.add(['a', 'a/b', 'a/b/c', 'a/d', 'a/d/e', 'f', 'f/g'])
        tree = self._convert_tree(work_tree)
        output_order = [p for p, e in tree.iter_entries_by_dir()]
        self.assertEqual(['', 'a', 'f', 'a/b', 'a/d', 'a/b/c', 'a/d/e', 'f/g'],
                         output_order)


class TestIterChildEntries(TestCaseWithTree):

    def test_iteration_order(self):
        work_tree = self.make_branch_and_tree('.')
        self.build_tree(['a/', 'a/b/', 'a/b/c', 'a/d/', 'a/d/e', 'f/', 'f/g'])
        work_tree.add(['a', 'a/b', 'a/b/c', 'a/d', 'a/d/e', 'f', 'f/g'])
        tree = self._convert_tree(work_tree)
        output = [e.name for e in tree.iter_child_entries('')]
        self.assertEqual({'a', 'f'}, set(output))
        output = [e.name for e in tree.iter_child_entries('a')]
        self.assertEqual({'b', 'd'}, set(output))

    def test_does_not_exist(self):
        work_tree = self.make_branch_and_tree('.')
        self.build_tree(['a/'])
        work_tree.add(['a'])
        tree = self._convert_tree(work_tree)
        self.assertRaises(_mod_transport.NoSuchFile, lambda:
                          list(tree.iter_child_entries('unknown')))


class TestExtras(TestCaseWithTree):

    def test_extras(self):
        work_tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/file', 'tree/versioned-file'])
        work_tree.add(['file', 'versioned-file'])
        work_tree.commit('add files')
        work_tree.remove('file')
        tree = self._convert_tree(work_tree)
        if isinstance(tree,
                      (revisiontree.RevisionTree,
                       workingtree_4.DirStateRevisionTree)):
            expected = []
        else:
            expected = ['file']
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(expected, list(tree.extras()))


class TestGetFileSha1(TestCaseWithTree):

    def test_get_file_sha1(self):
        work_tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([('tree/file', b'file content')])
        work_tree.add('file')
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        expected = osutils.sha_string(b'file content')
        self.assertEqual(expected, tree.get_file_sha1('file'))


class TestGetFileVerifier(TestCaseWithTree):

    def test_get_file_verifier(self):
        work_tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([
            ('tree/file1', b'file content'),
            ('tree/file2', b'file content')])
        work_tree.add(['file1', 'file2'])
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        (kind, data) = tree.get_file_verifier('file1')
        self.assertEqual(
            tree.get_file_verifier('file1'),
            tree.get_file_verifier('file2'))
        if kind == "SHA1":
            expected = osutils.sha_string(b'file content')
            self.assertEqual(expected, data)


class TestHasVersionedDirectories(TestCaseWithTree):

    def test_has_versioned_directories(self):
        work_tree = self.make_branch_and_tree('tree')
        tree = self._convert_tree(work_tree)
        self.assertIn(tree.has_versioned_directories(), (True, False))


class TestSupportsRenameTracking(TestCaseWithTree):

    def test_supports_rename_tracking(self):
        work_tree = self.make_branch_and_tree('tree')
        tree = self._convert_tree(work_tree)
        self.assertSubset([tree.supports_rename_tracking()], (True, False))


class TestSupportsVersionableKind(TestCaseWithTree):

    def test_file(self):
        work_tree = self.make_branch_and_tree('tree')
        tree = self._convert_tree(work_tree)
        self.assertTrue(tree.versionable_kind('file'))

    def test_unknown(self):
        work_tree = self.make_branch_and_tree('tree')
        tree = self._convert_tree(work_tree)
        self.assertFalse(tree.versionable_kind('unknown'))


class TestSpecialFilename(TestCaseWithTree):

    def test_is_special_path(self):
        work_tree = self.make_branch_and_tree('tree')
        tree = self._convert_tree(work_tree)
        self.assertFalse(tree.is_special_path('foo'))



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""A generator which creates a template-based output from the current
   tree info."""

import codecs

from breezy import errors
from breezy.lazy_regex import lazy_compile
from breezy.revision import NULL_REVISION
from breezy.version_info_formats import VersionInfoBuilder, create_date_str


class MissingTemplateVariable(errors.BzrError):

    _fmt = 'Variable {%(name)s} is not available.'

    def __init__(self, name):
        self.name = name


class NoTemplate(errors.BzrError):

    _fmt = 'No template specified.'


class Template:
    """A simple template engine.

    >>> t = Template()
    >>> t.add('test', 'xxx')
    >>> print(list(t.process('{test}')))
    ['xxx']
    >>> print(list(t.process('{test} test')))
    ['xxx', ' test']
    >>> print(list(t.process('test {test}')))
    ['test ', 'xxx']
    >>> print(list(t.process('test {test} test')))
    ['test ', 'xxx', ' test']
    >>> print(list(t.process('{test}\\\\n')))
    ['xxx', '\\n']
    >>> print(list(t.process('{test}\\n')))
    ['xxx', '\\n']
    """

    _tag_re = lazy_compile('{(\\w+)}')

    def __init__(self):
        self._data = {}

    def add(self, name, value):
        self._data[name] = value

    def process(self, tpl):
        unicode_escape = codecs.getdecoder("unicode_escape")
        tpl = unicode_escape(tpl)[0]
        pos = 0
        while True:
            match = self._tag_re.search(tpl, pos)
            if not match:
                if pos < len(tpl):
                    yield tpl[pos:]
                break
            start, end = match.span()
            if start > 0:
                yield tpl[pos:start]
            pos = end
            name = match.group(1)
            try:
                data = self._data[name]
            except KeyError:
                raise MissingTemplateVariable(name)
            if not isinstance(data, str):
                data = str(data)
            yield data


class CustomVersionInfoBuilder(VersionInfoBuilder):
    """Create a version file based on a custom template."""

    def generate(self, to_file):
        if self._template is None:
            raise NoTemplate()

        info = Template()
        info.add('build_date', create_date_str())
        info.add('branch_nick', self._branch.nick)

        revision_id = self._get_revision_id()
        if revision_id == NULL_REVISION:
            info.add('revno', 0)
        else:
            try:
                info.add('revno', self._get_revno_str(revision_id))
            except errors.GhostRevisionsHaveNoRevno:
                pass
            info.add('revision_id', revision_id.decode('utf-8'))
            rev = self._branch.repository.get_revision(revision_id)
            info.add('date', create_date_str(rev.timestamp, rev.timezone))

        if self._check:
            self._extract_file_revisions()

        if self._check:
            if self._clean:
                info.add('clean', 1)
            else:
                info.add('clean', 0)

        to_file.writelines(info.process(self._template))



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Branch.sprout()"""

import os

from breezy import branch as _mod_branch
from breezy import errors, osutils
from breezy import revision as _mod_revision
from breezy import tests, urlutils
from breezy.bzr import branch as _mod_bzrbranch
from breezy.bzr import remote
from breezy.tests import features
from breezy.tests.per_branch import TestCaseWithBranch


class TestSprout(TestCaseWithBranch):

    def test_sprout_branch_nickname(self):
        # test the nick name is reset always
        raise tests.TestSkipped('XXX branch sprouting is not yet tested.')

    def test_sprout_branch_parent(self):
        source = self.make_branch('source')
        target = source.controldir.sprout(self.get_url('target')).open_branch()
        self.assertEqual(
            urlutils.strip_segment_parameters(source.user_url),
            urlutils.strip_segment_parameters(target.get_parent()))

    def test_sprout_uses_bzrdir_branch_format(self):
        # branch.sprout(bzrdir) is defined as using the branch format selected
        # by bzrdir; format preservation is achieved by parameterising the
        # bzrdir during bzrdir.sprout, which is where stacking compatibility
        # checks are done. So this test tests that each implementation of
        # Branch.sprout delegates appropriately to the bzrdir which the
        # branch is being created in, rather than testing that the result is
        # in the format that we are testing (which is what would happen if
        # the branch did not delegate appropriately).
        if isinstance(self.branch_format, _mod_bzrbranch.BranchReferenceFormat):
            raise tests.TestNotApplicable('cannot sprout to a reference')
        # Start with a format that is unlikely to be the target format
        # We call the super class to allow overriding the format of creation)
        source = tests.TestCaseWithTransport.make_branch(self, 'old-branch',
                                                         format='knit')
        target_bzrdir = self.make_controldir('target')
        target_bzrdir.create_repository()
        result_format = self.branch_format
        if isinstance(target_bzrdir, remote.RemoteBzrDir):
            # for a remote bzrdir, we need to parameterise it with a branch
            # format, as, after creation, the newly opened remote objects
            # do not have one unless a branch was created at the time.
            # We use branch format 6 because its not the default, and its not
            # metaweave either.
            target_bzrdir._format.set_branch_format(
                _mod_bzrbranch.BzrBranchFormat6())
            result_format = target_bzrdir._format.get_branch_format()
        target = source.sprout(target_bzrdir)
        if isinstance(target, remote.RemoteBranch):
            # we have to look at the real branch to see whether RemoteBranch
            # did the right thing.
            target._ensure_real()
            target = target._real_branch
        if isinstance(result_format, remote.RemoteBranchFormat):
            # Unwrap a parameterised RemoteBranchFormat for comparison.
            result_format = result_format._custom_format
        self.assertIs(result_format.__class__, target._format.__class__)

    def test_sprout_partial(self):
        # test sprouting with a prefix of the revision-history.
        # also needs not-on-revision-history behaviour defined.
        wt_a = self.make_branch_and_tree('a')
        self.build_tree(['a/one'])
        wt_a.add(['one'])
        rev1 = wt_a.commit('commit one')
        self.build_tree(['a/two'])
        wt_a.add(['two'])
        wt_a.commit('commit two')
        repo_b = self.make_repository('b')
        repo_a = wt_a.branch.repository
        repo_a.copy_content_into(repo_b)
        br_b = wt_a.branch.sprout(repo_b.controldir, revision_id=rev1)
        self.assertEqual(rev1, br_b.last_revision())

    def test_sprout_partial_not_in_revision_history(self):
        """We should be able to sprout from any revision in ancestry."""
        wt = self.make_branch_and_tree('source')
        self.build_tree(['source/a'])
        wt.add('a')
        rev1 = wt.commit('rev1')
        rev2_alt = wt.commit('rev2-alt')
        wt.set_parent_ids([rev1])
        wt.branch.set_last_revision_info(1, rev1)
        rev2 = wt.commit('rev2')
        wt.set_parent_ids([rev2, rev2_alt])
        wt.commit('rev3')

        repo = self.make_repository('target')
        repo.fetch(wt.branch.repository)
        branch2 = wt.branch.sprout(repo.controldir, revision_id=rev2_alt)
        self.assertEqual((2, rev2_alt), branch2.last_revision_info())
        self.assertEqual(rev2_alt, branch2.last_revision())

    def test_sprout_preserves_tags(self):
        """Sprout preserves tags, even tags of absent revisions."""
        try:
            builder = self.make_branch_builder('source')
        except errors.UninitializableFormat:
            raise tests.TestSkipped('Uninitializable branch format')
        builder.build_commit(message="Rev 1")
        source = builder.get_branch()
        try:
            source.tags.set_tag('tag-a', b'missing-rev')
        except (errors.TagsNotSupported, errors.GhostTagsNotSupported):
            raise tests.TestNotApplicable(
                'Branch format does not support tags or tags to ghosts.')
        # Now source has a tag pointing to an absent revision.  Sprout it.
        target_bzrdir = self.make_repository('target').controldir
        new_branch = source.sprout(target_bzrdir)
        # The tag is present in the target
        self.assertEqual(b'missing-rev', new_branch.tags.lookup_tag('tag-a'))

    def test_sprout_from_any_repo_revision(self):
        """We should be able to sprout from any revision."""
        wt = self.make_branch_and_tree('source')
        self.build_tree(['source/a'])
        wt.add('a')
        rev1a = wt.commit('rev1a')
        # simulated uncommit
        wt.branch.set_last_revision_info(0, _mod_revision.NULL_REVISION)
        wt.set_last_revision(_mod_revision.NULL_REVISION)
        wt.revert()
        wt.commit('rev1b')
        wt2 = wt.controldir.sprout(
            'target', revision_id=rev1a).open_workingtree()
        self.assertEqual(rev1a, wt2.last_revision())
        self.assertPathExists('target/a')

    def test_sprout_with_unicode_symlink(self):
        # this tests bug #272444
        # Since the trigger function seems to be set_parent_trees, there exists
        # also a similar test, with name test_unicode_symlink, in class
        # TestSetParents at file per_workingtree/test_parents.py
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        self.requireFeature(features.UnicodeFilenameFeature)

        tree = self.make_branch_and_tree('tree1')

        # The link points to a file whose name is an omega
        # U+03A9 GREEK CAPITAL LETTER OMEGA
        # UTF-8: ce a9  UTF-16BE: 03a9  Decimal: &#937;
        target = '\u03a9'
        link_name = '\N{Euro Sign}link'
        os.symlink(target, 'tree1/' + link_name)
        tree.add([link_name])

        tree.commit('added a link to a Unicode target')
        tree.controldir.sprout('dest')
        self.assertEqual(target, osutils.readlink('dest/' + link_name))
        tree.lock_read()
        self.addCleanup(tree.unlock)
        # Check that the symlink target is safely round-tripped in the trees.
        self.assertEqual(target, tree.get_symlink_target(link_name))
        self.assertEqual(target,
                         tree.basis_tree().get_symlink_target(link_name))

    def test_sprout_with_ghost_in_mainline(self):
        tree = self.make_branch_and_tree('tree1')
        if not tree.branch.repository._format.supports_ghosts:
            raise tests.TestNotApplicable(
                "repository format does not support ghosts in mainline")
        tree.set_parent_ids([b"spooky"], allow_leftmost_as_ghost=True)
        tree.add('')
        rev1 = tree.commit('msg1')
        tree.commit('msg2')
        tree.controldir.sprout('target', revision_id=rev1)

    def assertBranchHookBranchIsStacked(self, pre_change_params):
        # Just calling will either succeed or fail.
        pre_change_params.branch.get_stacked_on_url()
        self.hook_calls.append(pre_change_params)

    def test_sprout_stacked_hooks_get_stacked_branch(self):
        tree = self.make_branch_and_tree('source')
        tree.commit('a commit')
        revid = tree.commit('a second commit')
        source = tree.branch
        target_transport = self.get_transport('target')
        self.hook_calls = []
        _mod_branch.Branch.hooks.install_named_hook("pre_change_branch_tip",
                                                    self.assertBranchHookBranchIsStacked, None)
        try:
            dir = source.controldir.sprout(target_transport.base,
                                           source.last_revision(), possible_transports=[target_transport],
                                           source_branch=source, stacked=True)
        except _mod_branch.UnstackableBranchFormat:
            if not self.branch_format.supports_stacking():
                raise tests.TestNotApplicable(
                    "Format doesn't auto stack successfully.")
            else:
                raise
        result = dir.open_branch()
        self.assertEqual(revid, result.last_revision())
        self.assertEqual(source.base, result.get_stacked_on_url())
        # Smart servers invoke hooks on both sides
        if isinstance(result, remote.RemoteBranch):
            expected_calls = 2
        else:
            expected_calls = 1
        self.assertEqual(expected_calls, len(self.hook_calls))



# Copyright (C) 2007, 2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for bisect_multi."""

from ..bisect_multi import bisect_multi_bytes
from . import TestCase


class TestBisectMultiBytes(TestCase):

    def test_lookup_no_keys_no_calls(self):
        calls = []

        def missing_content(location_keys):
            calls.append(location_keys)
            return ((location_key, False) for location_key in location_keys)
        self.assertEqual([],
                         list(bisect_multi_bytes(missing_content, 100, [])))
        self.assertEqual([], calls)

    def test_lookup_missing_key_no_content(self):
        """Doing a lookup in a zero-length file still does a single request.

        This makes sense because the bisector cannot tell how long content is
        and its more flexible to only stop when the content object says 'False'
        for a given location, key pair.
        """
        calls = []

        def missing_content(location_keys):
            calls.append(location_keys)
            return ((location_key, False) for location_key in location_keys)
        self.assertEqual([],
                         list(bisect_multi_bytes(missing_content, 0, ['foo', 'bar'])))
        self.assertEqual([[(0, 'foo'), (0, 'bar')]], calls)

    def test_lookup_missing_key_before_all_others(self):
        calls = []

        def missing_first_content(location_keys):
            # returns -1 for all keys unless the byte offset is 0 when it
            # returns False
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key[0] == 0:
                    result.append((location_key, False))
                else:
                    result.append((location_key, -1))
            return result
        # given a 0 length file, this should terminate with one call.
        self.assertEqual([],
                         list(bisect_multi_bytes(missing_first_content, 0, ['foo', 'bar'])))
        self.assertEqual([[(0, 'foo'), (0, 'bar')]], calls)
        del calls[:]
        # given a 2 length file, this should make two calls - 1, 0.
        self.assertEqual([],
                         list(bisect_multi_bytes(missing_first_content, 2, ['foo', 'bar'])))
        self.assertEqual([
            [(1, 'foo'), (1, 'bar')],
            [(0, 'foo'), (0, 'bar')],
            ], calls)
        del calls[:]
        # given a really long file - 200MB, this should make a series of calls with the
        # gap between adjactent calls dropping by 50% each time. We choose a
        # length which just under a power of two to generate a corner case in
        # bisection - naively using power of two reduction in size can lead to
        # a very long tail in the bisection process. The current users of
        # the bisect_multi_bytes api are not expected to be concerned by this,
        # as the delta gets down to 4K (the minimum we expect to read and
        # parse) within 16 steps even on a 200MB index (which at 4 keys/K is
        # 800 thousand keys, and log2 of 800000 is 19 - so we're doing log2
        # steps in the worst case there.
        self.assertEqual([],
                         list(bisect_multi_bytes(
                             missing_first_content, 268435456 - 1, ['foo', 'bar'])))
        self.assertEqual([
            [(134217727, 'foo'), (134217727, 'bar')],
            [(67108864, 'foo'), (67108864, 'bar')],
            [(33554433, 'foo'), (33554433, 'bar')],
            [(16777218, 'foo'), (16777218, 'bar')],
            [(8388611, 'foo'), (8388611, 'bar')],
            [(4194308, 'foo'), (4194308, 'bar')],
            [(2097157, 'foo'), (2097157, 'bar')],
            [(1048582, 'foo'), (1048582, 'bar')],
            [(524295, 'foo'), (524295, 'bar')],
            [(262152, 'foo'), (262152, 'bar')],
            [(131081, 'foo'), (131081, 'bar')],
            [(65546, 'foo'), (65546, 'bar')],
            [(32779, 'foo'), (32779, 'bar')],
            [(16396, 'foo'), (16396, 'bar')],
            [(8205, 'foo'), (8205, 'bar')],
            [(4110, 'foo'), (4110, 'bar')],
            [(2063, 'foo'), (2063, 'bar')],
            [(1040, 'foo'), (1040, 'bar')],
            [(529, 'foo'), (529, 'bar')],
            [(274, 'foo'), (274, 'bar')],
            [(147, 'foo'), (147, 'bar')],
            [(84, 'foo'), (84, 'bar')],
            [(53, 'foo'), (53, 'bar')],
            [(38, 'foo'), (38, 'bar')],
            [(31, 'foo'), (31, 'bar')],
            [(28, 'foo'), (28, 'bar')],
            [(27, 'foo'), (27, 'bar')],
            [(26, 'foo'), (26, 'bar')],
            [(25, 'foo'), (25, 'bar')],
            [(24, 'foo'), (24, 'bar')],
            [(23, 'foo'), (23, 'bar')],
            [(22, 'foo'), (22, 'bar')],
            [(21, 'foo'), (21, 'bar')],
            [(20, 'foo'), (20, 'bar')],
            [(19, 'foo'), (19, 'bar')],
            [(18, 'foo'), (18, 'bar')],
            [(17, 'foo'), (17, 'bar')],
            [(16, 'foo'), (16, 'bar')],
            [(15, 'foo'), (15, 'bar')],
            [(14, 'foo'), (14, 'bar')],
            [(13, 'foo'), (13, 'bar')],
            [(12, 'foo'), (12, 'bar')],
            [(11, 'foo'), (11, 'bar')],
            [(10, 'foo'), (10, 'bar')],
            [(9, 'foo'), (9, 'bar')],
            [(8, 'foo'), (8, 'bar')],
            [(7, 'foo'), (7, 'bar')],
            [(6, 'foo'), (6, 'bar')],
            [(5, 'foo'), (5, 'bar')],
            [(4, 'foo'), (4, 'bar')],
            [(3, 'foo'), (3, 'bar')],
            [(2, 'foo'), (2, 'bar')],
            [(1, 'foo'), (1, 'bar')],
            [(0, 'foo'), (0, 'bar')],
            ], calls)

    def test_lookup_missing_key_after_all_others(self):
        calls = []
        end = None

        def missing_last_content(location_keys):
            # returns +1 for all keys unless the byte offset is 'end' when it
            # returns False
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key[0] == end:
                    result.append((location_key, False))
                else:
                    result.append((location_key, +1))
            return result
        # given a 0 length file, this should terminate with one call.
        end = 0
        self.assertEqual([],
                         list(bisect_multi_bytes(missing_last_content, 0, ['foo', 'bar'])))
        self.assertEqual([[(0, 'foo'), (0, 'bar')]], calls)
        del calls[:]
        end = 2
        # given a 3 length file, this should make two calls - 1, 2.
        self.assertEqual([],
                         list(bisect_multi_bytes(missing_last_content, 3, ['foo', 'bar'])))
        self.assertEqual([
            [(1, 'foo'), (1, 'bar')],
            [(2, 'foo'), (2, 'bar')],
            ], calls)
        del calls[:]
        end = 268435456 - 2
        # see the really-big lookup series in
        # test_lookup_missing_key_before_all_others for details about this
        # assertion.
        self.assertEqual([],
                         list(bisect_multi_bytes(
                             missing_last_content, 268435456 - 1, ['foo', 'bar'])))
        self.assertEqual([
            [(134217727, 'foo'), (134217727, 'bar')],
            [(201326590, 'foo'), (201326590, 'bar')],
            [(234881021, 'foo'), (234881021, 'bar')],
            [(251658236, 'foo'), (251658236, 'bar')],
            [(260046843, 'foo'), (260046843, 'bar')],
            [(264241146, 'foo'), (264241146, 'bar')],
            [(266338297, 'foo'), (266338297, 'bar')],
            [(267386872, 'foo'), (267386872, 'bar')],
            [(267911159, 'foo'), (267911159, 'bar')],
            [(268173302, 'foo'), (268173302, 'bar')],
            [(268304373, 'foo'), (268304373, 'bar')],
            [(268369908, 'foo'), (268369908, 'bar')],
            [(268402675, 'foo'), (268402675, 'bar')],
            [(268419058, 'foo'), (268419058, 'bar')],
            [(268427249, 'foo'), (268427249, 'bar')],
            [(268431344, 'foo'), (268431344, 'bar')],
            [(268433391, 'foo'), (268433391, 'bar')],
            [(268434414, 'foo'), (268434414, 'bar')],
            [(268434925, 'foo'), (268434925, 'bar')],
            [(268435180, 'foo'), (268435180, 'bar')],
            [(268435307, 'foo'), (268435307, 'bar')],
            [(268435370, 'foo'), (268435370, 'bar')],
            [(268435401, 'foo'), (268435401, 'bar')],
            [(268435416, 'foo'), (268435416, 'bar')],
            [(268435423, 'foo'), (268435423, 'bar')],
            [(268435426, 'foo'), (268435426, 'bar')],
            [(268435427, 'foo'), (268435427, 'bar')],
            [(268435428, 'foo'), (268435428, 'bar')],
            [(268435429, 'foo'), (268435429, 'bar')],
            [(268435430, 'foo'), (268435430, 'bar')],
            [(268435431, 'foo'), (268435431, 'bar')],
            [(268435432, 'foo'), (268435432, 'bar')],
            [(268435433, 'foo'), (268435433, 'bar')],
            [(268435434, 'foo'), (268435434, 'bar')],
            [(268435435, 'foo'), (268435435, 'bar')],
            [(268435436, 'foo'), (268435436, 'bar')],
            [(268435437, 'foo'), (268435437, 'bar')],
            [(268435438, 'foo'), (268435438, 'bar')],
            [(268435439, 'foo'), (268435439, 'bar')],
            [(268435440, 'foo'), (268435440, 'bar')],
            [(268435441, 'foo'), (268435441, 'bar')],
            [(268435442, 'foo'), (268435442, 'bar')],
            [(268435443, 'foo'), (268435443, 'bar')],
            [(268435444, 'foo'), (268435444, 'bar')],
            [(268435445, 'foo'), (268435445, 'bar')],
            [(268435446, 'foo'), (268435446, 'bar')],
            [(268435447, 'foo'), (268435447, 'bar')],
            [(268435448, 'foo'), (268435448, 'bar')],
            [(268435449, 'foo'), (268435449, 'bar')],
            [(268435450, 'foo'), (268435450, 'bar')],
            [(268435451, 'foo'), (268435451, 'bar')],
            [(268435452, 'foo'), (268435452, 'bar')],
            [(268435453, 'foo'), (268435453, 'bar')],
            [(268435454, 'foo'), (268435454, 'bar')]
            ], calls)

    def test_lookup_when_a_key_is_missing_continues(self):
        calls = []

        def missing_foo_otherwise_missing_first_content(location_keys):
            # returns -1 for all keys unless the byte offset is 0 when it
            # returns False
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key[1] == 'foo' or location_key[0] == 0:
                    result.append((location_key, False))
                else:
                    result.append((location_key, -1))
            return result
        # given a 2 length file, this should terminate with two calls, one for
        # both keys, and one for bar only.
        self.assertEqual([],
                         list(bisect_multi_bytes(
                             missing_foo_otherwise_missing_first_content, 2,
                             ['foo', 'bar'])))
        self.assertEqual([
            [(1, 'foo'), (1, 'bar')],
            [(0, 'bar')],
            ], calls)

    def test_found_keys_returned_other_searches_continue(self):
        calls = []

        def find_bar_at_1_foo_missing_at_0(location_keys):
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key == (1, 'bar'):
                    result.append((location_key, 'bar-result'))
                elif location_key[0] == 0:
                    result.append((location_key, False))
                else:
                    result.append((location_key, -1))
            return result
        # given a 4 length file, this should terminate with three calls, two for
        # both keys, and one for foo only.
        self.assertEqual([('bar', 'bar-result')],
                         list(bisect_multi_bytes(
                             find_bar_at_1_foo_missing_at_0, 4,
                             ['foo', 'bar'])))
        self.assertEqual([
            [(2, 'foo'), (2, 'bar')],
            [(1, 'foo'), (1, 'bar')],
            [(0, 'foo')],
            ], calls)

    def test_searches_different_keys_in_different_directions(self):
        calls = []

        def missing_bar_at_1_foo_at_3(location_keys):
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key[1] == 'bar':
                    if location_key[0] == 1:
                        result.append((location_key, False))
                    else:
                        # search down
                        result.append((location_key, -1))
                elif location_key[1] == 'foo':
                    if location_key[0] == 3:
                        result.append((location_key, False))
                    else:
                        # search up
                        result.append((location_key, +1))
            return result
        # given a 4 length file, this should terminate with two calls.
        self.assertEqual([],
                         list(bisect_multi_bytes(
                             missing_bar_at_1_foo_at_3, 4,
                             ['foo', 'bar'])))
        self.assertEqual([
            [(2, 'foo'), (2, 'bar')],
            [(3, 'foo'), (1, 'bar')],
            ], calls)

    def test_change_direction_in_single_key_search(self):
        # check that we can search down, up, down again -
        # so length 8, goes 4, 6, 5
        calls = []

        def missing_at_5(location_keys):
            calls.append(location_keys)
            result = []
            for location_key in location_keys:
                if location_key[0] == 5:
                    result.append((location_key, False))
                elif location_key[0] > 5:
                    # search down
                    result.append((location_key, -1))
                else:
                    # search up
                    result.append((location_key, +1))
            return result
        # given a 8 length file, this should terminate with three calls.
        self.assertEqual([],
                         list(bisect_multi_bytes(
                             missing_at_5, 8,
                             ['foo', 'bar'])))
        self.assertEqual([
            [(4, 'foo'), (4, 'bar')],
            [(6, 'foo'), (6, 'bar')],
            [(5, 'foo'), (5, 'bar')],
            ], calls)



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from collections import deque
from io import BytesIO

from ... import debug, errors
from ...trace import mutter


class MessageHandler:
    """Base class for handling messages received via the smart protocol.

    As parts of a message are received, the corresponding PART_received method
    will be called.
    """

    def __init__(self):
        self.headers = None

    def headers_received(self, headers):
        """Called when message headers are received.

        This default implementation just stores them in self.headers.
        """
        self.headers = headers

    def byte_part_received(self, byte):
        """Called when a 'byte' part is received.

        Note that a 'byte' part is a message part consisting of exactly one
        byte.
        """
        raise NotImplementedError(self.byte_received)

    def bytes_part_received(self, bytes):
        """Called when a 'bytes' part is received.

        A 'bytes' message part can contain any number of bytes.  It should not
        be confused with a 'byte' part, which is always a single byte.
        """
        raise NotImplementedError(self.bytes_received)

    def structure_part_received(self, structure):
        """Called when a 'structure' part is received.

        :param structure: some structured data, which will be some combination
            of list, dict, int, and str objects.
        """
        raise NotImplementedError(self.bytes_received)

    def protocol_error(self, exception):
        """Called when there is a protocol decoding error.

        The default implementation just re-raises the exception.
        """
        raise

    def end_received(self):
        """Called when the end of the message is received."""
        # No-op by default.
        pass


class ConventionalRequestHandler(MessageHandler):
    """A message handler for "conventional" requests.

    "Conventional" is used in the sense described in
    doc/developers/network-protocol.txt: a simple message with arguments and an
    optional body.

    Possible states:
     * args: expecting args
     * body: expecting body (terminated by receiving a post-body status)
     * error: expecting post-body error
     * end: expecting end of message
     * nothing: finished
    """

    def __init__(self, request_handler, responder):
        MessageHandler.__init__(self)
        self.request_handler = request_handler
        self.responder = responder
        self.expecting = 'args'
        self._should_finish_body = False
        self._response_sent = False

    def protocol_error(self, exception):
        if self.responder.response_sent:
            # We can only send one response to a request, no matter how many
            # errors happen while processing it.
            return
        self.responder.send_error(exception)

    def byte_part_received(self, byte):
        if not isinstance(byte, bytes):
            raise TypeError(byte)
        if self.expecting == 'body':
            if byte == b'S':
                # Success.  Nothing more to come except the end of message.
                self.expecting = 'end'
            elif byte == b'E':
                # Error.  Expect an error structure.
                self.expecting = 'error'
            else:
                raise errors.SmartProtocolError(
                    'Non-success status byte in request body: {!r}'.format(byte))
        else:
            raise errors.SmartProtocolError(
                'Unexpected message part: byte({!r})'.format(byte))

    def structure_part_received(self, structure):
        if self.expecting == 'args':
            self._args_received(structure)
        elif self.expecting == 'error':
            self._error_received(structure)
        else:
            raise errors.SmartProtocolError(
                'Unexpected message part: structure({!r})'.format(structure))

    def _args_received(self, args):
        self.expecting = 'body'
        self.request_handler.args_received(args)
        if self.request_handler.finished_reading:
            self._response_sent = True
            self.responder.send_response(self.request_handler.response)
            self.expecting = 'end'

    def _error_received(self, error_args):
        self.expecting = 'end'
        self.request_handler.post_body_error_received(error_args)

    def bytes_part_received(self, bytes):
        if self.expecting == 'body':
            self._should_finish_body = True
            self.request_handler.accept_body(bytes)
        else:
            raise errors.SmartProtocolError(
                'Unexpected message part: bytes({!r})'.format(bytes))

    def end_received(self):
        if self.expecting not in ['body', 'end']:
            raise errors.SmartProtocolError(
                'End of message received prematurely (while expecting %s)'
                % (self.expecting,))
        self.expecting = 'nothing'
        self.request_handler.end_received()
        if not self.request_handler.finished_reading:
            raise errors.SmartProtocolError(
                "Complete conventional request was received, but request "
                "handler has not finished reading.")
        if not self._response_sent:
            self.responder.send_response(self.request_handler.response)


class ResponseHandler:
    """Abstract base class for an object that handles a smart response."""

    def read_response_tuple(self, expect_body=False):
        """Reads and returns the response tuple for the current request.

        :keyword expect_body: a boolean indicating if a body is expected in the
            response.  Some protocol versions needs this information to know
            when a response is finished.  If False, read_body_bytes should
            *not* be called afterwards.  Defaults to False.
        :returns: tuple of response arguments.
        """
        raise NotImplementedError(self.read_response_tuple)

    def read_body_bytes(self, count=-1):
        """Read and return some bytes from the body.

        :param count: if specified, read up to this many bytes.  By default,
            reads the entire body.
        :returns: str of bytes from the response body.
        """
        raise NotImplementedError(self.read_body_bytes)

    def read_streamed_body(self):
        """Returns an iterable that reads and returns a series of body chunks.
        """
        raise NotImplementedError(self.read_streamed_body)

    def cancel_read_body(self):
        """Stop expecting a body for this response.

        If expect_body was passed to read_response_tuple, this cancels that
        expectation (and thus finishes reading the response, allowing a new
        request to be issued).  This is useful if a response turns out to be an
        error rather than a normal result with a body.
        """
        raise NotImplementedError(self.cancel_read_body)


class ConventionalResponseHandler(MessageHandler, ResponseHandler):

    def __init__(self):
        MessageHandler.__init__(self)
        self.status = None
        self.args = None
        self._bytes_parts = deque()
        self._body_started = False
        self._body_stream_status = None
        self._body = None
        self._body_error_args = None
        self.finished_reading = False

    def setProtoAndMediumRequest(self, protocol_decoder, medium_request):
        self._protocol_decoder = protocol_decoder
        self._medium_request = medium_request

    def byte_part_received(self, byte):
        if not isinstance(byte, bytes):
            raise TypeError(byte)
        if byte not in [b'E', b'S']:
            raise errors.SmartProtocolError(
                'Unknown response status: {!r}'.format(byte))
        if self._body_started:
            if self._body_stream_status is not None:
                raise errors.SmartProtocolError(
                    'Unexpected byte part received: {!r}'.format(byte))
            self._body_stream_status = byte
        else:
            if self.status is not None:
                raise errors.SmartProtocolError(
                    'Unexpected byte part received: {!r}'.format(byte))
            self.status = byte

    def bytes_part_received(self, bytes):
        self._body_started = True
        self._bytes_parts.append(bytes)

    def structure_part_received(self, structure):
        if not isinstance(structure, tuple):
            raise errors.SmartProtocolError(
                'Args structure is not a sequence: {!r}'.format(structure))
        if not self._body_started:
            if self.args is not None:
                raise errors.SmartProtocolError(
                    'Unexpected structure received: %r (already got %r)'
                    % (structure, self.args))
            self.args = structure
        else:
            if self._body_stream_status != b'E':
                raise errors.SmartProtocolError(
                    'Unexpected structure received after body: %r'
                    % (structure,))
            self._body_error_args = structure

    def _wait_for_response_args(self):
        while self.args is None and not self.finished_reading:
            self._read_more()

    def _wait_for_response_end(self):
        while not self.finished_reading:
            self._read_more()

    def _read_more(self):
        next_read_size = self._protocol_decoder.next_read_size()
        if next_read_size == 0:
            # a complete request has been read.
            self.finished_reading = True
            self._medium_request.finished_reading()
            return
        data = self._medium_request.read_bytes(next_read_size)
        if data == b'':
            # end of file encountered reading from server
            if 'hpss' in debug.debug_flags:
                mutter(
                    'decoder state: buf[:10]=%r, state_accept=%s',
                    self._protocol_decoder._get_in_buffer()[:10],
                    self._protocol_decoder.state_accept.__name__)
            raise errors.ConnectionReset(
                "Unexpected end of message. "
                "Please check connectivity and permissions, and report a bug "
                "if problems persist.")
        self._protocol_decoder.accept_bytes(data)

    def protocol_error(self, exception):
        # Whatever the error is, we're done with this request.
        self.finished_reading = True
        self._medium_request.finished_reading()
        raise

    def read_response_tuple(self, expect_body=False):
        """Read a response tuple from the wire."""
        self._wait_for_response_args()
        if not expect_body:
            self._wait_for_response_end()
        if 'hpss' in debug.debug_flags:
            mutter('   result:   %r', self.args)
        if self.status == b'E':
            self._wait_for_response_end()
            _raise_smart_server_error(self.args)
        return tuple(self.args)

    def read_body_bytes(self, count=-1):
        """Read bytes from the body, decoding into a byte stream.

        We read all bytes at once to ensure we've checked the trailer for
        errors, and then feed the buffer back as read_body_bytes is called.

        Like the builtin file.read in Python, a count of -1 (the default) means
        read the entire body.
        """
        # TODO: we don't necessarily need to buffer the full request if count
        # != -1.  (2008/04/30, Andrew Bennetts)
        if self._body is None:
            self._wait_for_response_end()
            body_bytes = b''.join(self._bytes_parts)
            if 'hpss' in debug.debug_flags:
                mutter('              %d body bytes read', len(body_bytes))
            self._body = BytesIO(body_bytes)
            self._bytes_parts = None
        return self._body.read(count)

    def read_streamed_body(self):
        while not self.finished_reading:
            while self._bytes_parts:
                bytes_part = self._bytes_parts.popleft()
                if 'hpssdetail' in debug.debug_flags:
                    mutter('              %d byte part read', len(bytes_part))
                yield bytes_part
            self._read_more()
        if self._body_stream_status == b'E':
            _raise_smart_server_error(self._body_error_args)

    def cancel_read_body(self):
        self._wait_for_response_end()


def _raise_smart_server_error(error_tuple):
    """Raise exception based on tuple received from smart server

    Specific error translation is handled by breezy.bzr.remote._translate_error
    """
    if error_tuple[0] == b'UnknownMethod':
        raise errors.UnknownSmartMethod(error_tuple[1])
    raise errors.ErrorFromSmartServer(error_tuple)



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import config
from breezy.errors import SSHVendorNotFound, UnknownSSH
from breezy.tests import TestCase, TestCaseWithTransport
from breezy.transport.ssh import (LSHSubprocessVendor, OpenSSHSubprocessVendor,
                                  PLinkSubprocessVendor,
                                  SSHCorpSubprocessVendor, SSHVendorManager,
                                  StrangeHostname)


class TestSSHVendorManager(SSHVendorManager):

    _ssh_version_string = ""

    def set_ssh_version_string(self, version):
        self._ssh_version_string = version

    def _get_ssh_version_string(self, args):
        return self._ssh_version_string


class SSHVendorManagerTests(TestCaseWithTransport):

    def test_register_vendor(self):
        manager = TestSSHVendorManager()
        self.overrideEnv('BRZ_SSH', None)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)
        vendor = object()
        manager.register_vendor("vendor", vendor)
        self.overrideEnv('BRZ_SSH', 'vendor')
        self.assertIs(manager.get_vendor(), vendor)

    def test_default_vendor(self):
        manager = TestSSHVendorManager()
        self.overrideEnv('BRZ_SSH', None)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)
        vendor = object()
        manager.register_default_vendor(vendor)
        self.assertIs(manager.get_vendor(), vendor)

    def test_get_vendor_by_environment(self):
        manager = TestSSHVendorManager()
        self.overrideEnv('BRZ_SSH', None)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)
        self.overrideEnv('BRZ_SSH', 'vendor')
        self.assertRaises(UnknownSSH, manager.get_vendor)
        vendor = object()
        manager.register_vendor("vendor", vendor)
        self.assertIs(manager.get_vendor(), vendor)

    def test_get_vendor_by_config(self):
        manager = TestSSHVendorManager()
        self.overrideEnv('BRZ_SSH', None)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)
        config.GlobalStack().set('ssh', 'vendor')
        self.assertRaises(UnknownSSH, manager.get_vendor)
        vendor = object()
        manager.register_vendor("vendor", vendor)
        self.assertIs(manager.get_vendor(), vendor)

    def test_get_vendor_by_inspection_openssh(self):
        manager = TestSSHVendorManager()
        self.overrideEnv('BRZ_SSH', None)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)
        manager.set_ssh_version_string("OpenSSH")
        self.assertIsInstance(manager.get_vendor(), OpenSSHSubprocessVendor)

    def test_get_vendor_by_inspection_sshcorp(self):
        manager = TestSSHVendorManager()
        self.overrideEnv('BRZ_SSH', None)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)
        manager.set_ssh_version_string("SSH Secure Shell")
        self.assertIsInstance(manager.get_vendor(), SSHCorpSubprocessVendor)

    def test_get_vendor_by_inspection_lsh(self):
        manager = TestSSHVendorManager()
        self.overrideEnv('BRZ_SSH', None)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)
        manager.set_ssh_version_string("lsh")
        self.assertIsInstance(manager.get_vendor(), LSHSubprocessVendor)

    def test_get_vendor_by_inspection_plink(self):
        manager = TestSSHVendorManager()
        self.overrideEnv('BRZ_SSH', None)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)
        manager.set_ssh_version_string("plink")
        # Auto-detect of plink vendor disabled, on Windows recommended
        # default ssh-client is paramiko
        # see https://bugs.launchpad.net/bugs/414743
        #~self.assertIsInstance(manager.get_vendor(), PLinkSubprocessVendor)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)

    def test_cached_vendor(self):
        manager = TestSSHVendorManager()
        self.overrideEnv('BRZ_SSH', None)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)
        vendor = object()
        manager.register_vendor("vendor", vendor)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)
        # Once the vendor is found the result is cached (mainly because of the
        # 'get_vendor' sometimes can be an expensive operation) and later
        # invocations of the 'get_vendor' just returns the cached value.
        self.overrideEnv('BRZ_SSH', 'vendor')
        self.assertIs(manager.get_vendor(), vendor)
        self.overrideEnv('BRZ_SSH', None)
        self.assertIs(manager.get_vendor(), vendor)
        # The cache can be cleared by the 'clear_cache' method
        manager.clear_cache()
        self.overrideEnv('BRZ_SSH', None)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)

    def test_get_vendor_search_order(self):
        # The 'get_vendor' method search for SSH vendors as following:
        #
        #   1. Check previously cached value
        #   2. Check BRZ_SSH environment variable
        #   3. Check the system for known SSH vendors
        #   4. Fall back to the default vendor if registered
        #
        # Let's now check the each check method in the reverse order
        # clearing the cache between each invocation:

        manager = TestSSHVendorManager()
        # At first no vendors are found
        self.overrideEnv('BRZ_SSH', None)
        self.assertRaises(SSHVendorNotFound, manager.get_vendor)

        # If the default vendor is registered it will be returned
        default_vendor = object()
        manager.register_default_vendor(default_vendor)
        self.assertIs(manager.get_vendor(), default_vendor)

        # If the known vendor is found in the system it will be returned
        manager.clear_cache()
        manager.set_ssh_version_string("OpenSSH")
        self.assertIsInstance(manager.get_vendor(), OpenSSHSubprocessVendor)

        # If the BRZ_SSH environment variable is found it will be treated as
        # the vendor name
        manager.clear_cache()
        vendor = object()
        manager.register_vendor("vendor", vendor)
        self.overrideEnv('BRZ_SSH', 'vendor')
        self.assertIs(manager.get_vendor(), vendor)

        # Last cached value always checked first
        self.overrideEnv('BRZ_SSH', 'vendor')
        self.assertIs(manager.get_vendor(), vendor)

    def test_get_vendor_from_path_win32_plink(self):
        manager = TestSSHVendorManager()
        manager.set_ssh_version_string("plink: Release 0.60")
        plink_path = "C:/Program Files/PuTTY/plink.exe"
        self.overrideEnv('BRZ_SSH', plink_path)
        vendor = manager.get_vendor()
        self.assertIsInstance(vendor, PLinkSubprocessVendor)
        args = vendor._get_vendor_specific_argv("user", "host", 22, ["bzr"])
        self.assertEqual(args[0], plink_path)

    def test_get_vendor_from_path_nix_openssh(self):
        manager = TestSSHVendorManager()
        manager.set_ssh_version_string(
            "OpenSSH_5.1p1 Debian-5, OpenSSL, 0.9.8g 19 Oct 2007")
        openssh_path = "/usr/bin/ssh"
        self.overrideEnv('BRZ_SSH', openssh_path)
        vendor = manager.get_vendor()
        self.assertIsInstance(vendor, OpenSSHSubprocessVendor)
        args = vendor._get_vendor_specific_argv("user", "host", 22, ["bzr"])
        self.assertEqual(args[0], openssh_path)


class SubprocessVendorsTests(TestCase):

    def test_openssh_command_tricked(self):
        vendor = OpenSSHSubprocessVendor()
        self.assertEqual(
            vendor._get_vendor_specific_argv(
                "user", "-oProxyCommand=blah", 100, command=["bzr"]),
            ["ssh", "-oForwardX11=no", "-oForwardAgent=no",
                "-oClearAllForwardings=yes",
                "-oNoHostAuthenticationForLocalhost=yes",
                "-p", "100",
                "-l", "user",
                "--",
                "-oProxyCommand=blah", "bzr"])

    def test_openssh_command_arguments(self):
        vendor = OpenSSHSubprocessVendor()
        self.assertEqual(
            vendor._get_vendor_specific_argv(
                "user", "host", 100, command=["bzr"]),
            ["ssh", "-oForwardX11=no", "-oForwardAgent=no",
                "-oClearAllForwardings=yes",
                "-oNoHostAuthenticationForLocalhost=yes",
                "-p", "100",
                "-l", "user",
                "--",
                "host", "bzr"]
            )

    def test_openssh_subsystem_arguments(self):
        vendor = OpenSSHSubprocessVendor()
        self.assertEqual(
            vendor._get_vendor_specific_argv(
                "user", "host", 100, subsystem="sftp"),
            ["ssh", "-oForwardX11=no", "-oForwardAgent=no",
                "-oClearAllForwardings=yes",
                "-oNoHostAuthenticationForLocalhost=yes",
                "-p", "100",
                "-l", "user",
                "-s", "--", "host", "sftp"]
            )

    def test_openssh_command_strange_hostname(self):
        vendor = SSHCorpSubprocessVendor()
        self.assertRaises(
            StrangeHostname,
            vendor._get_vendor_specific_argv,
            "user", "-oProxyCommand=host", 100, command=["bzr"])

    def test_sshcorp_command_arguments(self):
        vendor = SSHCorpSubprocessVendor()
        self.assertEqual(
            vendor._get_vendor_specific_argv(
                "user", "host", 100, command=["bzr"]),
            ["ssh", "-x",
                "-p", "100",
                "-l", "user",
                "host", "bzr"]
            )

    def test_sshcorp_subsystem_arguments(self):
        vendor = SSHCorpSubprocessVendor()
        self.assertEqual(
            vendor._get_vendor_specific_argv(
                "user", "host", 100, subsystem="sftp"),
            ["ssh", "-x",
                "-p", "100",
                "-l", "user",
                "-s", "sftp", "host"]
            )

    def test_lsh_command_tricked(self):
        vendor = LSHSubprocessVendor()
        self.assertRaises(
            StrangeHostname,
            vendor._get_vendor_specific_argv,
            "user", "-oProxyCommand=host", 100, command=["bzr"])

    def test_lsh_command_arguments(self):
        vendor = LSHSubprocessVendor()
        self.assertEqual(
            vendor._get_vendor_specific_argv(
                "user", "host", 100, command=["bzr"]),
            ["lsh",
                "-p", "100",
                "-l", "user",
                "host", "bzr"]
            )

    def test_lsh_subsystem_arguments(self):
        vendor = LSHSubprocessVendor()
        self.assertEqual(
            vendor._get_vendor_specific_argv(
                "user", "host", 100, subsystem="sftp"),
            ["lsh",
                "-p", "100",
                "-l", "user",
                "--subsystem", "sftp", "host"]
            )

    def test_plink_command_tricked(self):
        vendor = PLinkSubprocessVendor()
        self.assertRaises(
            StrangeHostname,
            vendor._get_vendor_specific_argv,
            "user", "-oProxyCommand=host", 100, command=["bzr"])

    def test_plink_command_arguments(self):
        vendor = PLinkSubprocessVendor()
        self.assertEqual(
            vendor._get_vendor_specific_argv(
                "user", "host", 100, command=["bzr"]),
            ["plink", "-x", "-a", "-ssh", "-2", "-batch",
                "-P", "100",
                "-l", "user",
                "host", "bzr"]
            )

    def test_plink_subsystem_arguments(self):
        vendor = PLinkSubprocessVendor()
        self.assertEqual(
            vendor._get_vendor_specific_argv(
                "user", "host", 100, subsystem="sftp"),
            ["plink", "-x", "-a", "-ssh", "-2", "-batch",
                "-P", "100",
                "-l", "user",
                "-s", "host", "sftp"]
            )



# Copyright (C) 2019 Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License or
# (at your option) a later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Fossil foreign branch support.

Currently only tells the user that Fossil is not supported.
"""

from ... import version_info  # noqa: F401
from ... import controldir, errors


class FossilUnsupportedError(errors.UnsupportedVcs):

    vcs = "fossil"

    _fmt = ('Fossil branches are not yet supported. '
            'To interoperate with Fossil branches, use fastimport.')


class FossilDirFormat(controldir.ControlDirFormat):
    """Fossil directory format."""

    def get_converter(self):
        raise NotImplementedError(self.get_converter)

    def get_format_description(self):
        return "Fossil control directory"

    def initialize_on_transport(self, transport):
        raise errors.UninitializableFormat(format=self)

    def is_supported(self):
        return False

    def supports_transport(self, transport):
        return False

    def check_support_status(self, allow_unsupported, recommend_upgrade=True,
                             basedir=None):
        raise FossilUnsupportedError(format=self)

    def open(self, transport):
        # Raise NotBranchError if there is nothing there
        RemoteFossilProber().probe_transport(transport)
        raise NotImplementedError(self.open)


class RemoteFossilProber(controldir.Prober):

    @classmethod
    def priority(klass, transport):
        return 95

    @classmethod
    def probe_transport(klass, transport):
        from breezy.transport.http.urllib import HttpTransport
        if not isinstance(transport, HttpTransport):
            raise errors.NotBranchError(path=transport.base)
        response = transport.request(
            'POST', transport.base, headers={'Content-Type': 'application/x-fossil'})
        if response.status == 501:
            raise errors.NotBranchError(path=transport.base)
        ct = response.getheader('Content-Type')
        if ct is None:
            raise errors.NotBranchError(path=transport.base)
        if ct.split(';')[0] != 'application/x-fossil':
            raise errors.NotBranchError(path=transport.base)
        return FossilDirFormat()

    @classmethod
    def known_formats(cls):
        return [FossilDirFormat()]


controldir.ControlDirFormat.register_prober(RemoteFossilProber)
